import difflib
import sys
import types
from dataclasses import dataclass, fields, field
from inspect import isclass
from itertools import chain
from typing import Any, Type, TypedDict, Union, get_args, get_origin

import pyparsing as pp

from ..enums import AccountEdition, DataType, ParseableEnum, ResourceType
from ..identifiers import FQN, URN, parse_identifier, resource_label_for_type
from ..lifecycle import create_resource, drop_resource
from ..parse import _parse_create_header, _parse_props, _resolve_resource_class
from ..props import Props as ResourceProps
from ..resource_name import ResourceName
from ..resource_tags import ResourceTags
from ..scope import (
    AccountScope,
    DatabaseScope,
    OrganizationScope,
    ResourceScope,
    SchemaScope,
    resource_can_be_contained_in,
)


class WrongContainerException(Exception):
    pass


class ResourceHasContainerException(Exception):
    pass


def _suggest_correct_kwargs(expected_kwargs, passed_kwargs):
    suggestions = {}
    for passed_kwarg in passed_kwargs:
        # Find the closest match from the expected kwargs for each passed kwarg
        closest_match = difflib.get_close_matches(passed_kwarg, expected_kwargs, n=1, cutoff=0.6)
        if closest_match:
            closest_match = closest_match[0]
            # If the passed kwarg is not exactly an expected kwarg, add it to suggestions
            if passed_kwarg != closest_match:
                suggestions[passed_kwarg] = closest_match
        else:
            # If no close match is found, suggest it might be an unexpected kwarg
            suggestions[passed_kwarg] = "Unexpected kwarg, no close match found."

    return suggestions


class Arg(TypedDict):
    name: str
    data_type: DataType


class Returns(TypedDict):
    data_type: DataType
    metadata: str


@dataclass
class LifecycleConfig:
    ignore_changes: list[str] = field(default_factory=list)
    prevent_destroy: bool = False


def _coerce_resource_field(field_value, field_type):

    # No type checking or coercion for Any
    if field_type == Any:
        return field_value
    elif field_type == str:
        if not isinstance(field_value, str):
            raise TypeError
        return field_value

    # Recursively traverse lists and dicts
    if get_origin(field_type) == list:
        if not isinstance(field_value, list):
            raise TypeError
        list_element_type = get_args(field_type) or (str,)
        return [_coerce_resource_field(v, field_type=list_element_type[0]) for v in field_value]
    elif get_origin(field_type) == dict:
        if not isinstance(field_value, dict):
            raise TypeError
        dict_types = get_args(field_type)
        if len(dict_types) < 2:
            raise RuntimeError(f"Unexpected field type {field_type}")
        return {k: _coerce_resource_field(v, field_type=dict_types[1]) for k, v in field_value.items()}

    # Check for field_value's type in a Union
    if get_origin(field_type) == Union:
        union_types = get_args(field_type)
        for union_type in union_types:
            expected_type = get_origin(union_type) or union_type
            if isinstance(field_value, expected_type):
                return _coerce_resource_field(field_value, field_type=expected_type)
        raise RuntimeError(f"Unexpected field type {field_type}")

    if isinstance(field_type, str):
        # Special case for Tag owner which has the type "Role" due to a circular import issue
        for subclass in Resource.__subclasses__():
            if subclass.__name__ == field_type:
                field_type = subclass
                break
        else:
            raise RuntimeError(f"Class for type {field_type} not found")

    if not isclass(field_type):
        # If we want to support quoted type annotations, this is the place to do it.
        # eg owner "Role" = "SYSADMIN"
        raise RuntimeError(f"Unexpected field type {field_type}")

    # Coerce enums
    if issubclass(field_type, ParseableEnum):
        return field_type(field_value)

    # Coerce args
    elif field_type == Arg:
        arg_dict = {
            "name": field_value["name"].upper(),
            "data_type": DataType(field_value["data_type"]),
        }
        if "default" in field_value:
            arg_dict["default"] = field_value["default"]
        return arg_dict

    # Coerce returns
    elif field_type == Returns:
        returns_dict = {
            "data_type": DataType(field_value["data_type"]),
            "metadata": field_value["metadata"],
        }
        if "returns_null" in field_value:
            returns_dict["returns_null"] = field_value["returns_null"]
        return returns_dict

    # Coerce resources
    elif issubclass(field_type, Resource):
        return convert_to_resource(field_type, field_value)
    elif field_type == ResourceName:
        return ResourceName(field_value)
    elif field_type == ResourceTags:
        return ResourceTags(field_value)
    else:
        return field_value


@dataclass
class ResourceSpec:
    def __post_init__(self):
        for f in fields(self):
            field_value = getattr(self, f.name)
            if field_value is None:
                continue
            else:
                try:
                    new_value = _coerce_resource_field(field_value, f.type)
                    setattr(self, f.name, new_value)
                except TypeError as err:
                    human_readable_classname = self.__class__.__name__[1:]
                    raise TypeError(
                        f"Expected {human_readable_classname}.{f.name} to be {f.type}, got {repr(field_value)} instead"
                    ) from err

    @classmethod
    def get_metadata(cls, field_name: str):
        return {f.name: f.metadata for f in fields(cls)}[field_name]


RESOURCE_SCOPES = {
    ResourceType.ACCOUNT: OrganizationScope(),
}


class _Resource(type):
    __types__: dict[ResourceType, list[Type["Resource"]]] = {}
    __resolvers__ = {}

    def __new__(cls, name, bases, attrs):
        cls_ = super().__new__(cls, name, bases, attrs)
        if cls_.__name__ in ["Resource", "_Resource", "ResourcePointer"]:
            return cls_
        if cls_.resource_type not in cls.__types__:
            cls.__types__[cls_.resource_type] = []
        cls.__types__[cls_.resource_type].append(cls_)
        if cls_.resource_type not in RESOURCE_SCOPES:
            RESOURCE_SCOPES[cls_.resource_type] = cls_.scope
        return cls_


class Resource(metaclass=_Resource):
    edition = {AccountEdition.STANDARD, AccountEdition.ENTERPRISE, AccountEdition.BUSINESS_CRITICAL}
    props: ResourceProps
    resource_type: ResourceType
    scope: ResourceScope
    spec: Type[ResourceSpec]
    serialize_inline: bool = False

    def __init__(
        self,
        implicit: bool = False,
        lifecycle: dict = None,
        **kwargs,
    ):
        super().__init__()
        self._data: ResourceSpec = None
        self._container: "ResourceContainer" = None
        self._finalized = False
        self.lifecycle = LifecycleConfig(**lifecycle) if lifecycle else LifecycleConfig()
        self.implicit = implicit
        self.refs: set[Resource] = set()

        # Consume resource_type from kwargs if it exists
        resource_type = kwargs.pop("resource_type", None)
        resource_type = ResourceType(resource_type) if resource_type else None
        if resource_type and resource_type != self.resource_type:
            raise ValueError(f"Unexpected resource_type {resource_type} for {self.resource_type}")

        # Consume scope from kwargs if it exists
        database = kwargs.pop("database", None)
        schema = kwargs.pop("schema", None)
        self._register_scope(database=database, schema=schema)

        # If there are more kwargs, throw an error
        # Based on https://stackoverflow.com/questions/1603940/how-can-i-modify-a-python-traceback-object-when-raising-an-exception
        if kwargs:
            try:
                if self.spec:
                    field_names = [f.name for f in fields(self.spec)]
                    field_names = ", ".join(field_names)
                    raise ValueError(
                        f"Unexpected keyword arguments for {self.__class__.__name__} {kwargs}. Valid field names: {field_names}"
                    )
                else:
                    raise ValueError(f"Unexpected keyword arguments for {self.__class__.__name__} {kwargs}")
            except ValueError as err:
                traceback = sys.exc_info()[2]
                back_frame = traceback.tb_frame.f_back
                msg = str(err)

            back_tb = types.TracebackType(
                tb_next=None, tb_frame=back_frame, tb_lasti=back_frame.f_lasti, tb_lineno=back_frame.f_lineno
            )
            raise ValueError(msg).with_traceback(back_tb)

    @classmethod
    def from_sql(cls, sql):
        resource_cls = cls
        if resource_cls == Resource:
            # FIXME: we need to change the way we handle polymorphic resources
            # make a new function called _parse_resource_type_from_create
            # resource_cls = Resource.classes[_resolve_resource_class(sql)]
            # raise NotImplementedError
            resource_type = _resolve_resource_class(sql)
            scope = RESOURCE_SCOPES[resource_type]
        else:
            resource_type = resource_cls.resource_type
            scope = resource_cls.scope

        identifier, remainder_sql = _parse_create_header(sql, resource_type, scope)

        try:
            props = _parse_props(resource_cls.props, remainder_sql) if remainder_sql else {}
            return resource_cls(**identifier, **props)
        except pp.ParseException as err:
            raise pp.ParseException(f"Error parsing {resource_cls.__name__} props {identifier}") from err

    @classmethod
    def from_dict(cls, data: dict):
        resource_cls = cls.resolve_resource_cls(ResourceType(data["resource_type"]), data)
        return resource_cls(**data)

    @classmethod
    def props_for_resource_type(cls, resource_type: ResourceType, data: dict = None):
        return cls.resolve_resource_cls(resource_type, data).props

    @classmethod
    def resolve_resource_cls(cls, resource_type: ResourceType, data: dict = None) -> Type["Resource"]:

        if not isinstance(resource_type, ResourceType):
            raise ValueError(f"Expected ResourceType, got {resource_type}({type(resource_type)})")
        if resource_type not in cls.__types__:
            raise ValueError(f"Resource class for type not found {resource_type}")
        resource_types = cls.__types__[resource_type]

        if len(resource_types) > 1:
            if data is None:
                raise ValueError(f"Cannot resolve polymorphic resource class [{resource_type}] without data")
            else:
                resolver = cls.__resolvers__[resource_type]
                return resolver(data)
        return resource_types[0]

    @classmethod
    def defaults(cls):
        return {f.name: f.default for f in fields(cls.spec)}

    def __repr__(self):  # pragma: no cover
        name = getattr(self._data, "name", None)
        implicit = "~" if self.implicit else ""
        return f"{self.__class__.__name__}({implicit}{name})"

    def __eq__(self, other):
        if not isinstance(other, Resource):
            return False
        return self._data == other._data

    def __hash__(self):
        return hash(URN.from_resource(self, ""))

    def to_dict(self):
        serialized = {}

        def _serialize(value):
            if isinstance(value, ResourcePointer):
                return str(value.fqn)
            elif isinstance(value, Resource):
                # if hasattr(value, "serialize"):
                #     return value.serialize()
                if getattr(value, "serialize_inline", False):
                    return value.to_dict()
                if hasattr(value._data, "name"):
                    return getattr(value._data, "name")
                else:
                    raise Exception(f"Cannot serialize {value}")
            elif isinstance(value, ParseableEnum):
                return str(value)
            elif isinstance(value, list):
                return [_serialize(v) for v in value]
            elif isinstance(value, dict):
                return {k: _serialize(v) for k, v in value.items()}
            elif isinstance(value, ResourceName):
                return str(value)
            elif isinstance(value, ResourceTags):
                return value.tags
            else:
                return value

        # for key, value in asdict(self._data).items():
        #     serialized[key] = _serialize(value)
        for f in fields(self._data):
            value = getattr(self._data, f.name)
            serialized[f.name] = _serialize(value)

        return serialized

    def create_sql(self, **kwargs):
        return create_resource(
            self.urn,
            self.to_dict(),
            self.props,
            **kwargs,
        )

    def drop_sql(self, if_exists: bool = False):
        return drop_resource(self.urn, self.to_dict(), if_exists=if_exists)

    def _requires(self, resource: "Resource"):
        if self._finalized:
            raise RuntimeError("Cannot modify a finalized resource")
        if isinstance(resource, (Resource, ResourcePointer)):
            self.refs.add(resource)

    def requires(self, *resources: "Resource"):
        if isinstance(resources[0], list):
            resources = resources[0]
        for resource in resources:
            self._requires(resource)

    def _register_scope(self, database=None, schema=None):
        if isinstance(database, str):
            database: ResourceContainer = ResourcePointer(name=database, resource_type=ResourceType.DATABASE)

        if isinstance(schema, str):
            schema: ResourceContainer = ResourcePointer(name=schema, resource_type=ResourceType.SCHEMA)
            if database is not None:
                database.add(schema)

        if isinstance(self.scope, DatabaseScope):
            if schema is not None:
                raise RuntimeError(f"Unexpected kwarg schema {schema} for resource {self}")
            if database is not None:
                database.add(self)

        if isinstance(self.scope, SchemaScope):
            if schema is not None:
                schema.add(self)
            elif database is not None:
                database.find(name="PUBLIC", resource_type=ResourceType.SCHEMA).add(self)

    def to_pointer(self):
        return ResourcePointer(
            name=str(self.fqn),
            resource_type=self.resource_type,
        )

    @property
    def container(self):
        return self._container

    @property
    def urn(self):
        return URN.from_resource(self, account_locator="")

    @property
    def fqn(self) -> FQN:
        raise NotImplementedError("Subclasses must implement fqn")


class ResourceContainer:
    def __init__(self):
        self._items: dict[ResourceType, list[Resource]] = {}

    def __contains__(self, item: Resource):
        return item in self._items.get(item.resource_type, [])

    def add(self, *items: Resource):
        if isinstance(items[0], list):
            items = items[0]
        for item in items:
            if not resource_can_be_contained_in(item, self):
                raise WrongContainerException(f"{item} cannot be added to {self}")
            if item.container is not None and not isinstance(item.container, ResourcePointer):
                raise ResourceHasContainerException(f"{item} already belongs to a container")
            item._container = self
            item.requires(self)
            if item.resource_type not in self._items:
                self._items[item.resource_type] = []
            self._items[item.resource_type].append(item)

    def items(self, resource_type: ResourceType = None) -> list[Resource]:
        if resource_type:
            return self._items.get(resource_type, [])
        else:
            return list(chain.from_iterable(self._items.values()))

    def find(self, resource_type: ResourceType, name: str) -> Resource:
        for resource in self.items(resource_type):
            if isinstance(resource, ResourcePointer) and resource.name == name:
                return resource
            elif (
                isinstance(resource, Resource)
                and resource._data is not None
                and getattr(resource._data, "name", None) == name
            ):
                return resource
        raise KeyError(f"Resource {resource_type} {name} not found")

    def remove(self, resource: Resource):
        if resource.resource_type in self._items:
            for index, item in enumerate(self._items[resource.resource_type]):
                if item is resource:
                    self._items[resource.resource_type].pop(index)
                    resource._container = None
            resource.refs.remove(self)


class NamedResource:
    """
    This class is a mixin that allows resources to be constructed with fully qualified names
    without a bunch of user-land cruft to make it all work. This keeps titan close to the behavior
    of snowflake.

    For example:

    Without this trait, a scoped resource would need to be defined as:
    >>> tbl = Table(name="TBL", database="DB", schema="SCHEMA")

    With this trait, the same table can be defined this way instead:
    >>> tbl = Table(name="DB.SCHEMA.TBL")
    """

    def __init__(self, name: str, **kwargs):
        name = str(name)

        if isinstance(self.scope, AccountScope):
            self._name = ResourceName(name)
            super().__init__(**kwargs)
            return

        try:
            identifier = parse_identifier(name, is_db_scoped=isinstance(self.scope, DatabaseScope))
        except pp.ParseException:
            # Allow identifiers that should be quoted, so long as they aren't insane
            if "." in name:
                raise ValueError(f"Resource name not supported {name}")
            identifier = parse_identifier(f'"{name}"', is_db_scoped=isinstance(self.scope, DatabaseScope))

        self._name = ResourceName(identifier["name"])
        if "database" in identifier and "database" in kwargs:
            raise ValueError("Multiple database names found")
        if "schema" in identifier and "schema" in kwargs:
            raise ValueError("Multiple schema names found")

        super().__init__(
            database=kwargs.pop("database", identifier.get("database")),
            schema=kwargs.pop("schema", identifier.get("schema")),
            **kwargs,
        )

    @property
    def name(self):
        return self._name

    @property
    def fqn(self) -> FQN:
        return self.scope.fully_qualified_name(self.container, self.name)


class ResourcePointer(NamedResource, Resource, ResourceContainer):
    def __init__(self, name: str, resource_type: ResourceType):
        self._resource_type: ResourceType = resource_type
        self.scope = RESOURCE_SCOPES[resource_type]
        super().__init__(name)

        # Don't want to do this for all implicit resources but making an exception for PUBLIC schema
        # If this points to a database, assume it includes a PUBLIC schema
        if self._resource_type == ResourceType.DATABASE and self._name != "SNOWFLAKE":
            self.add(ResourcePointer(name="PUBLIC", resource_type=ResourceType.SCHEMA))
            # self.add(ResourcePointer(name="INFORMATION_SCHEMA", resource_type=ResourceType.SCHEMA))

    def __repr__(self):  # pragma: no cover
        resource_type = getattr(self, "resource_type", None)
        resource_label = resource_label_for_type(resource_type).title() if resource_type else "resource"
        name = getattr(self, "name", None)
        return f"{resource_label}(→{name})"

    def __eq__(self, other):
        if not isinstance(other, ResourcePointer):
            return False
        return self.name == other.name and self.resource_type == other.resource_type

    def __hash__(self):
        return hash((self._name, self._resource_type))

    @property
    def container(self):
        return self._container

    @property
    def fqn(self):
        return self.scope.fully_qualified_name(self.container, self.name)

    @property
    def resource_type(self):
        return self._resource_type

    def to_dict(self):
        return {
            "_pointer": True,
            "name": self.name,
        }


def convert_to_resource(cls: Resource, resource_or_descriptor: Union[str, dict, Resource, ResourceName]) -> Resource:
    """Convert a resource descriptor to a resource instance"""
    if isinstance(resource_or_descriptor, str) or isinstance(resource_or_descriptor, ResourceName):
        return ResourcePointer(name=resource_or_descriptor, resource_type=cls.resource_type)
    elif isinstance(resource_or_descriptor, dict):
        # We permit two types of anonymous resource descriptors
        # 1. A dict with a single key, the name of the resource
        # Example:
        #   grant = Grant(priv='SELECT', on='table', to={'name': 'some_role'})
        # 2. A dict representing a column or column-like object
        # Example:
        #   table = Table(name='my_table', columns=[{'name': 'id', 'data_type': 'int'}])

        # if cls.__name__ == "Column":
        #     return cls(**resource_or_descriptor)
        if cls.serialize_inline:
            return cls(**resource_or_descriptor)
        else:
            return ResourcePointer(**resource_or_descriptor, resource_type=cls.resource_type)
    elif isinstance(resource_or_descriptor, cls):
        return resource_or_descriptor
    elif isinstance(resource_or_descriptor, ResourcePointer):
        if resource_or_descriptor.resource_type == cls.resource_type:
            return resource_or_descriptor
        else:
            raise ValueError(f"Unexpected resource type {resource_or_descriptor}")
    else:
        raise ValueError(f"Unexpected resource descriptor {resource_or_descriptor}")



from dataclasses import dataclass, field

# from .stage import InternalStage, copy_options
from ..enums import ResourceType
from ..parse import _parse_create_header, _parse_props, _parse_table_schema
from ..props import (
    BoolProp,
    FlagProp,
    IdentifierListProp,
    IntProp,
    Props,
    SchemaProp,
    StringProp,
    TagsProp,
)
from ..scope import SchemaScope
from .column import Column
from .resource import NamedResource, Resource, ResourceName, ResourceSpec
from .role import Role
from .tag import TaggableResource


@dataclass(unsafe_hash=True)
class _Table(ResourceSpec):
    name: ResourceName
    # TODO: allow columns to be specified as SQL
    columns: list[Column]
    constraints: list[str] = None
    transient: bool = False
    cluster_by: list[str] = None
    enable_schema_evolution: bool = False
    data_retention_time_in_days: int = None
    max_data_extension_time_in_days: int = None
    change_tracking: bool = False
    default_ddl_collation: str = None
    copy_grants: bool = field(default_factory=None, metadata={"fetchable": False})
    row_access_policy: dict[str, list] = None
    owner: Role = "SYSADMIN"
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.columns is None:
            raise ValueError("columns can't be None")
        if len(self.columns) == 0:
            raise ValueError("columns can't be empty")


class Table(NamedResource, TaggableResource, Resource):
    """
    Description:
        A table in Snowflake.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-table

    Fields:
        name (string, required): The name of the table.
        columns (list, required): The columns of the table.
        constraints (list): The constraints of the table.
        transient (bool): Whether the table is transient.
        cluster_by (list): The clustering keys for the table.
        enable_schema_evolution (bool): Whether schema evolution is enabled. Defaults to False.
        data_retention_time_in_days (int): The data retention time in days.
        max_data_extension_time_in_days (int): The maximum data extension time in days.
        change_tracking (bool): Whether change tracking is enabled. Defaults to False.
        default_ddl_collation (string): The default DDL collation.
        copy_grants (bool): Whether to copy grants. Defaults to False.
        row_access_policy (dict): The row access policy.
        tags (dict): The tags for the table.
        owner (string or Role): The owner role of the table. Defaults to SYSADMIN.
        comment (string): A comment for the table.

    Python:

        ```python
        table = Table(
            name="some_table",
            columns=[{"name": "col1", "data_type": "STRING"}],
            owner="SYSADMIN",
        )
        ```

    Yaml:

        ```yaml
        tables:
          - name: some_table
            columns:
              - name: col1
                data_type: STRING
            owner: SYSADMIN
        ```
    """

    resource_type = ResourceType.TABLE
    props = Props(
        transient=FlagProp("transient"),
        columns=SchemaProp(),
        cluster_by=IdentifierListProp("cluster by", eq=False, parens=True),
        enable_schema_evolution=BoolProp("enable_schema_evolution"),
        data_retention_time_in_days=IntProp("data_retention_time_in_days"),
        max_data_extension_time_in_days=IntProp("max_data_extension_time_in_days"),
        change_tracking=BoolProp("change_tracking"),
        default_ddl_collation=StringProp("default_ddl_collation"),
        copy_grants=FlagProp("copy grants"),
        tags=TagsProp(),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _Table

    def __init__(
        self,
        name: str,
        columns: list[Column],
        constraints: list[str] = None,
        transient: bool = False,
        cluster_by: list[str] = None,
        enable_schema_evolution: bool = False,
        data_retention_time_in_days: int = None,
        max_data_extension_time_in_days: int = None,
        change_tracking: bool = False,
        default_ddl_collation: str = None,
        copy_grants: bool = None,
        row_access_policy: dict[str, list] = None,
        tags: dict[str, str] = None,
        owner: str = "SYSADMIN",
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data = _Table(
            name=self._name,
            columns=columns,
            constraints=constraints,
            transient=transient,
            cluster_by=cluster_by,
            enable_schema_evolution=enable_schema_evolution,
            data_retention_time_in_days=data_retention_time_in_days,
            max_data_extension_time_in_days=max_data_extension_time_in_days,
            change_tracking=change_tracking,
            default_ddl_collation=default_ddl_collation,
            copy_grants=copy_grants,
            row_access_policy=row_access_policy,
            owner=owner,
            comment=comment,
        )
        self._table_stage = None
        self.set_tags(tags)
        # self._table_stage: InternalStage = InternalStage(name=f"@%{self.name}", implicit=True)
        # if self.schema:
        #     self._table_stage.schema = self.schema

    @classmethod
    def from_sql(cls, sql):
        """
        CREATE [ OR REPLACE ]
        [ { [ { LOCAL | GLOBAL } ] TEMP | TEMPORARY | VOLATILE | TRANSIENT } ]
        TABLE [ IF NOT EXISTS ] <table_name>
        ( ... )
        [ CLUSTER BY ( <expr> [ , <expr> , ... ] ) ]
        [ ENABLE_SCHEMA_EVOLUTION = { TRUE | FALSE } ]
        [ STAGE_FILE_FORMAT = ( { FORMAT_NAME = '<file_format_name>'
                                | TYPE = { CSV | JSON | AVRO | ORC | PARQUET | XML } [ formatTypeOptions ] } ) ]
        [ STAGE_COPY_OPTIONS = ( copyOptions ) ]
        [ DATA_RETENTION_TIME_IN_DAYS = <integer> ]
        [ MAX_DATA_EXTENSION_TIME_IN_DAYS = <integer> ]
        [ CHANGE_TRACKING = { TRUE | FALSE } ]
        [ DEFAULT_DDL_COLLATION = '<collation_specification>' ]
        [ COPY GRANTS ]
        [ [ WITH ] ROW ACCESS POLICY <policy_name> ON ( <col_name> [ , <col_name> ... ] ) ]
        [ [ WITH ] TAG ( <tag_name> = '<tag_value>' [ , <tag_name> = '<tag_value>' , ... ] ) ]
        [ COMMENT = '<string_literal>' ]
        """

        identifier, remainder = _parse_create_header(sql, cls.resource_type, cls.scope)
        table_schema, remainder = _parse_table_schema(remainder)
        props = _parse_props(cls.props, remainder)
        return cls(**identifier, **table_schema, **props)

    @property
    def table_stage(self):
        return self._table_stage


# @dataclass(unsafe_hash=True)
# class _CreateTableAsSelect(ResourceSpec):
#     name: str
#     as_: str = field(default_factory=None, metadata={"triggers_replacement": True})
#     columns: list[Column] = field(default_factory=None, metadata={"triggers_replacement": True})
#     cluster_by: list[str] = None
#     copy_grants: bool = False
#     row_access_policy: dict[str, list] = None
#     owner: Role = "SYSADMIN"
#     comment: str = None

#     def __post_init__(self):
#         super().__post_init__()
#         if self.as_ is None:
#             raise ValueError("as can't be None")


# class CreateTableAsSelect(Resource):
#     resource_type = ResourceType.TABLE
#     props = Props(
#         columns=SchemaProp(),
#         cluster_by=IdentifierListProp("cluster by", eq=False, parens=True),
#         copy_grants=FlagProp("copy grants"),
#         as_=QueryProp("as"),
#     )
#     scope = SchemaScope()
#     spec = _CreateTableAsSelect

#     def __init__(
#         self,
#         name: str,
#         columns: list[Column] = None,
#         cluster_by: list[str] = None,
#         copy_grants: bool = False,
#         row_access_policy: dict[str, list] = None,
#         as_: str = None,
#         owner: str = "SYSADMIN",
#         **kwargs,
#     ):
#         super().__init__(**kwargs)
#         self._data = _CreateTableAsSelect(
#             name=name,
#             as_=as_,
#             columns=columns,
#             cluster_by=cluster_by,
#             copy_grants=copy_grants,
#             row_access_policy=row_access_policy,
#             owner=owner,
#         )

#     @classmethod
#     def from_sql(cls, sql):
#         """
#         CREATE [ OR REPLACE ] TABLE <table_name> [ ( <col_name> [ <col_type> ] , <col_name> [ <col_type> ] , ... ) ]
#         [ CLUSTER BY ( <expr> [ , <expr> , ... ] ) ]
#         [ COPY GRANTS ]
#         AS SELECT <query>
#         [ ... ]

#         """

#         identifier, remainder = _parse_create_header(sql, cls.resource_type, cls.scope)
#         table_schema, remainder = _parse_table_schema(remainder)
#         props = _parse_props(cls.props, remainder)
#         return cls(**identifier, **table_schema, **props)


# # there's no detectable difference between a table created with CTAS and a table created with a regular CREATE TABLE
# def _resolver(data: dict):
#     if "as_" in data:
#         return CreateTableAsSelect
#     return Table


# Resource.__resolvers__[ResourceType.TABLE] = _resolver



from dataclasses import dataclass, field

from ..enums import AccountEdition, ResourceType
from ..props import (
    BoolProp,
    IdentifierListProp,
    IdentifierProp,
    IntProp,
    Props,
    StringProp,
    TagsProp,
)
from ..resource_name import ResourceName
from ..scope import SchemaScope
from .compute_pool import ComputePool
from .resource import NamedResource, Resource, ResourceSpec
from .tag import TaggableResource
from .warehouse import Warehouse


@dataclass(unsafe_hash=True)
class _ServiceSpec(ResourceSpec):
    name: ResourceName
    compute_pool: ComputePool
    stage: str = field(default_factory=None, metadata={"fetchable": False})
    yaml_file_stage_path: str = field(default_factory=None, metadata={"fetchable": False})
    specification: str = field(default_factory=None, metadata={"fetchable": False})
    external_access_integrations: list[str] = None
    auto_resume: bool = True
    min_instances: int = None
    max_instances: int = None
    query_warehouse: Warehouse = None
    comment: str = None


class Service(NamedResource, TaggableResource, Resource):
    """
    Description:
        Service is a managed resource in Snowflake that allows users to run instances of their applications
        as a collection of containers on a compute pool. Each service instance can handle incoming traffic
        with the help of a load balancer if multiple instances are run.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-service

    Fields:
        name (string, required): The unique identifier for the service within the schema.
        compute_pool (string or ComputePool, required): The compute pool on which the service runs.
        stage (string): The Snowflake internal stage where the specification file is stored.
        yaml_file_stage_path (string): The path to the service specification file on the stage.
        specification (string): The service specification as a string.
        external_access_integrations (list): The names of external access integrations for the service.
        auto_resume (bool): Specifies whether to automatically resume the service when a function or ingress is called. Defaults to True.
        min_instances (int): The minimum number of service instances to run.
        max_instances (int): The maximum number of service instances to run.
        query_warehouse (string or Warehouse): The warehouse to use if a service container connects to Snowflake to execute a query.
        tags (dict): Tags associated with the service.
        comment (string): A comment for the service.

    Python:

        ```python
        service = Service(
            name="some_service",
            compute_pool="some_compute_pool",
            stage="@tutorial_stage",
            yaml_file_stage_path="echo_spec.yaml",
            specification="FROM SPECIFICATION $$some_specification$$",
            external_access_integrations=["some_integration"],
            auto_resume=True,
            min_instances=1,
            max_instances=2,
            query_warehouse="some_warehouse",
            tags={"key": "value"},
            comment="This is a sample service."
        )
        ```

    Yaml:

        ```yaml
        services:
          - name: some_service
            compute_pool: some_compute_pool
            stage: @tutorial_stage
            yaml_file_stage_path: echo_spec.yaml
            specification: FROM SPECIFICATION $$some_specification$$
            external_access_integrations:
              - some_integration
            auto_resume: true
            min_instances: 1
            max_instances: 2
            query_warehouse: some_warehouse
            tags:
              key: value
            comment: This is a sample service.
        ```
    """

    edition = {AccountEdition.ENTERPRISE, AccountEdition.BUSINESS_CRITICAL}
    resource_type = ResourceType.SERVICE
    props = Props(
        compute_pool=IdentifierProp("in compute pool", eq=False),
        specification_file=StringProp("specification_file"),
        specification=StringProp("from specification", eq=False),
        external_access_integrations=IdentifierListProp("external_access_integrations", parens=True),
        auto_resume=BoolProp("auto_resume"),
        min_instances=IntProp("min_instances"),
        max_instances=IntProp("max_instances"),
        query_warehouse=IdentifierProp("query_warehouse"),
        tags=TagsProp(),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _ServiceSpec

    def __init__(
        self,
        name: str,
        compute_pool: ComputePool,
        stage: str = None,
        yaml_file_stage_path: str = None,
        specification: str = None,
        external_access_integrations: list[str] = None,
        auto_resume: bool = True,
        min_instances: int = None,
        max_instances: int = None,
        query_warehouse: Warehouse = None,
        tags: dict[str, str] = None,
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _ServiceSpec = _ServiceSpec(
            name=self._name,
            compute_pool=compute_pool,
            stage=stage,
            yaml_file_stage_path=yaml_file_stage_path,
            specification=specification,
            external_access_integrations=external_access_integrations,
            auto_resume=auto_resume,
            min_instances=min_instances,
            max_instances=max_instances,
            query_warehouse=query_warehouse,
            comment=comment,
        )
        self.set_tags(tags)



import yaml

from functools import cache

from titan.data_provider import list_schemas
from titan.enums import ResourceType
from titan.identifiers import FQN
from titan.privs import DatabasePriv, TablePriv, SchemaPriv, ViewPriv, WarehousePriv
from titan.resources import GrantOnAll, FutureGrant, Grant, RoleGrant
from titan.resources.resource import ResourcePointer


DATABASE_READ_PRIVS = [DatabasePriv.USAGE]
DATABASE_WRITE_PRIVS = [DatabasePriv.USAGE, DatabasePriv.MONITOR, DatabasePriv.CREATE_SCHEMA]

STORAGE_INTEGRATION_PRIVS = ["USAGE"]

SCHEMA_READ_PRIVS = [SchemaPriv.USAGE]
SCHEMA_WRITE_PRIVS = [
    SchemaPriv.USAGE,
    SchemaPriv.MONITOR,
    SchemaPriv.CREATE_ALERT,
    SchemaPriv.CREATE_DYNAMIC_TABLE,
    SchemaPriv.CREATE_EXTERNAL_TABLE,
    SchemaPriv.CREATE_FILE_FORMAT,
    SchemaPriv.CREATE_FUNCTION,
    # SchemaPriv.CREATE_MASKING_POLICY,
    # SchemaPriv.CREATE_MATERIALIZED_VIEW,
    # SchemaPriv.CREATE_NETWORK_RULE,
    # SchemaPriv.CREATE_PACKAGES_POLICY,
    # SchemaPriv.CREATE_PASSWORD_POLICY,
    SchemaPriv.CREATE_PIPE,
    SchemaPriv.CREATE_PROCEDURE,
    # SchemaPriv.CREATE_ROW_ACCESS_POLICY,
    SchemaPriv.CREATE_SECRET,
    SchemaPriv.CREATE_SEQUENCE,
    # SchemaPriv.CREATE_SESSION_POLICY,
    # SchemaPriv.CREATE_SNOWFLAKE_ML_ANOMALY_DETECTION,
    # SchemaPriv.CREATE_SNOWFLAKE_ML_FORECAST,
    SchemaPriv.CREATE_STAGE,
    SchemaPriv.CREATE_STREAM,
    SchemaPriv.CREATE_TABLE,
    # SchemaPriv.CREATE_TAG,
    SchemaPriv.CREATE_TASK,
    SchemaPriv.CREATE_VIEW,
]

TABLE_READ_PRIVS = [TablePriv.SELECT]
TABLE_WRITE_PRIVS = [
    TablePriv.SELECT,
    TablePriv.INSERT,
    TablePriv.UPDATE,
    TablePriv.DELETE,
    TablePriv.TRUNCATE,
    TablePriv.REFERENCES,
]

VIEW_READ_PRIVS = [ViewPriv.SELECT]
VIEW_WRITE_PRIVS = [ViewPriv.SELECT]

WAREHOUSE_PRIVS = [WarehousePriv.USAGE, WarehousePriv.OPERATE, WarehousePriv.MONITOR]


def _parse_permifrost_identifier(identifier: str, is_db_scoped: bool = False):
    """
    Parse a permifrost identifier into a database and schema.

    Args:
        identifier (str): The identifier to parse.
        is_db_scoped (bool): Whether the identifier is scoped to a database.

    Returns:
        tuple: A tuple containing the database and schema.
    """
    parts = identifier.split(".")
    if is_db_scoped:
        return FQN(database=parts[0], name=parts[1])
    return FQN(database=parts[0], schema=parts[1], name=parts[2])


@cache
def _list_schemas(session, database):
    return list_schemas(session, database)


def read_permifrost_config(session, file_path):
    """
    Read the permifrost config file and return the config as a dictionary.

    Args:
        file_path (str): The path to the permifrost config file.

    Returns:
        dict: The permifrost config as a dictionary.
    """
    # return resources.read_yaml(file_path)
    config = {}
    with open(file_path, "r") as file:
        config = yaml.safe_load(file)

    version = config.pop("version", None)
    databases = config.pop("databases", [])
    roles = config.pop("roles", [])
    users = config.pop("users", [])
    warehouses = config.pop("warehouses", [])
    integrations = config.pop("integrations", [])

    return [
        # *databases,
        *_get_role_resources(session, roles),
        *_get_user_resources(users),
        # warehouses,
        # integrations,
    ]


def _get_role_resources(session, roles: list):
    resources = []
    for permifrost_role in roles:
        role_name, config = permifrost_role.popitem()
        role = ResourcePointer(name=role_name, resource_type=ResourceType.ROLE)
        resources.append(role)

        warehouses = config.get("warehouses", [])
        for wh in warehouses:
            resources.append(ResourcePointer(name=wh, resource_type=ResourceType.WAREHOUSE))
            for priv in WAREHOUSE_PRIVS:
                resources.append(Grant(priv=priv, on_warehouse=wh, to=role))

        integrations = config.get("integrations", [])
        for integration in integrations:
            resources.append(ResourcePointer(name=integration, resource_type=ResourceType.STORAGE_INTEGRATION))
            for priv in STORAGE_INTEGRATION_PRIVS:
                resources.append(Grant(priv=priv, on_storage_integration=integration, to=role))

        member_of = config.get("member_of", [])
        for parent_role in member_of:
            if parent_role == "*":
                continue
            resources.append(RoleGrant(role=role, to_role=parent_role))

        database_read = config.get("privileges", {}).get("databases", {}).get("read", [])
        database_write = config.get("privileges", {}).get("databases", {}).get("write", [])

        def _add_database_grants(resources, databases, privs, role):
            for db in databases:
                if db.upper() == "SNOWFLAKE":
                    continue
                resources.append(ResourcePointer(name=db, resource_type=ResourceType.DATABASE))
                for priv in privs:
                    resources.append(Grant(priv=priv, on_database=db, to=role))

        _add_database_grants(resources, database_read, DATABASE_READ_PRIVS, role)
        _add_database_grants(resources, database_write, DATABASE_WRITE_PRIVS, role)

        schema_read = config.get("privileges", {}).get("schemas", {}).get("read", [])
        schema_write = config.get("privileges", {}).get("schemas", {}).get("write", [])

        def _add_schema_grants(resources, schema_identifier, privs, role):
            if schema_identifier.endswith(".*"):
                database = _parse_permifrost_identifier(schema_identifier, is_db_scoped=True).database
                for schema in _list_schemas(session, database):
                    if schema.endswith("INFORMATION_SCHEMA"):
                        continue
                    for priv in privs:
                        resources.append(Grant(priv=priv, on_schema=schema, to=role))
                for priv in privs:
                    resources.append(FutureGrant(priv=priv, on_future_schemas_in_database=database, to=role))
            elif schema_identifier.endswith("*"):
                # schema: "db.schema_*"
                return
            else:
                fqn = _parse_permifrost_identifier(schema_identifier, is_db_scoped=True)
                if fqn.database.upper() == "SNOWFLAKE":
                    return
                db = ResourcePointer(name=fqn.database, resource_type=ResourceType.DATABASE)
                schema = ResourcePointer(name=fqn.name, resource_type=ResourceType.SCHEMA)
                db.add(schema)
                resources.append(db)
                resources.append(schema)
                for priv in privs:
                    resources.append(Grant(priv=priv, on_schema=schema_identifier, to=role))

        for schema in schema_read:
            _add_schema_grants(resources, schema, SCHEMA_READ_PRIVS, role)
        for schema in schema_write:
            _add_schema_grants(resources, schema, SCHEMA_WRITE_PRIVS, role)

        table_read = config.get("privileges", {}).get("tables", {}).get("read", [])
        table_write = config.get("privileges", {}).get("tables", {}).get("write", [])

        def _add_table_grants(resources, table_identifier, privs, role):
            if table_identifier.endswith(".*.*"):
                database = _parse_permifrost_identifier(table_identifier).database
                for schema in _list_schemas(session, database):
                    if schema.endswith("INFORMATION_SCHEMA"):
                        continue
                    for priv in privs:
                        resources.append(GrantOnAll(priv=priv, on_all_tables_in_schema=schema, to=role))
                        resources.append(FutureGrant(priv=priv, on_future_tables_in_schema=schema, to=role))

                for priv in privs:
                    resources.append(GrantOnAll(priv=priv, on_all_tables_in_database=database, to=role))
                    resources.append(FutureGrant(priv=priv, on_future_tables_in_database=database, to=role))
            elif table_identifier.endswith(".*"):
                fqn = _parse_permifrost_identifier(table_identifier)
                if fqn.schema.endswith("*") or fqn.database.upper() == "SNOWFLAKE":
                    return
                db = ResourcePointer(name=fqn.database, resource_type=ResourceType.DATABASE)
                schema = ResourcePointer(name=fqn.schema, resource_type=ResourceType.SCHEMA)
                db.add(schema)
                resources.append(db)
                resources.append(schema)
                for priv in privs:
                    resources.append(GrantOnAll(priv=priv, on_all_tables_in=schema, to=role))
                    resources.append(FutureGrant(priv=priv, on_future_tables_in=schema, to=role))
            elif table_identifier.endswith("*"):
                # table: "db.schema.table_*"
                return
            else:
                resources.append(ResourcePointer(name=table_identifier, resource_type=ResourceType.TABLE))
                for priv in privs:
                    resources.append(Grant(priv=priv, on_table=table_identifier, to=role))

        def _add_view_grants(resources, view_identifier, privs, role):
            if view_identifier.endswith(".*.*"):
                database = _parse_permifrost_identifier(view_identifier).database
                for schema in _list_schemas(session, database):
                    if schema.endswith("INFORMATION_SCHEMA"):
                        continue
                    for priv in privs:
                        resources.append(GrantOnAll(priv=priv, on_all_views_in_schema=schema, to=role))
                        resources.append(FutureGrant(priv=priv, on_future_views_in_schema=schema, to=role))
                for priv in privs:
                    resources.append(GrantOnAll(priv=priv, on_all_views_in_database=database, to=role))
                    resources.append(FutureGrant(priv=priv, on_future_views_in_database=database, to=role))
            elif view_identifier.endswith(".*"):
                fqn = _parse_permifrost_identifier(view_identifier)
                if fqn.schema.endswith("*") or fqn.database.upper() == "SNOWFLAKE":
                    return
                db = ResourcePointer(name=fqn.database, resource_type=ResourceType.DATABASE)
                schema = ResourcePointer(name=fqn.schema, resource_type=ResourceType.SCHEMA)
                db.add(schema)
                resources.append(db)
                resources.append(schema)
                for priv in privs:
                    resources.append(GrantOnAll(priv=priv, on_all_views_in=schema, to=role))
                    resources.append(FutureGrant(priv=priv, on_future_views_in=schema, to=role))

            elif view_identifier.endswith("*"):
                # table: "db.schema.table_*"
                return
            else:
                resources.append(ResourcePointer(name=view_identifier, resource_type=ResourceType.VIEW))
                for priv in privs:
                    resources.append(Grant(priv=priv, on_view=view_identifier, to=role))

        for table in table_read:
            _add_table_grants(resources, table, TABLE_READ_PRIVS, role)
            _add_view_grants(resources, table, VIEW_READ_PRIVS, role)
        for table in table_write:
            _add_table_grants(resources, table, TABLE_WRITE_PRIVS, role)
            _add_view_grants(resources, table, VIEW_WRITE_PRIVS, role)

        # TODO: owns

    return resources


def _get_user_resources(users: list):
    resources = []
    for permifrost_user in users:
        user, config = permifrost_user.popitem()
        resources.append(ResourcePointer(name=user, resource_type=ResourceType.USER))

        member_of = config.get("member_of", [])
        for role in member_of:
            if role == "*":
                continue
            resources.append(RoleGrant(role=role, to_user=user))

    return resources



# Stored Procedure Interface (spi)
import inspect
import pydoc
import re
import sys

from snowflake.snowpark.exceptions import SnowparkSQLException

from . import lifecycle, resources
from . import data_provider as dp
from .blueprint import Action, Blueprint
from .diff import diff
from .enums import DataType, ResourceType
from .identifiers import FQN, URN, parse_FQN
from .scope import DatabaseScope, SchemaScope

SNOWPARK_TELEMETRY_ID = "titan_titan"

try:
    import _snowflake  # type: ignore

    _snowflake.snowflake_partner_attribution().append(SNOWPARK_TELEMETRY_ID)
except ModuleNotFoundError as err:
    # raise ModuleNotFoundError("The titan spi module can only be run from a Snowpark UDF or stored procedure") from err
    pass

__this__ = sys.modules[__name__]

_PYTHON_TO_SNOWFLAKE_TYPE = {
    int: DataType.INTEGER,
    float: DataType.FLOAT,
    str: DataType.VARCHAR,
    bool: DataType.BOOLEAN,
    dict: DataType.OBJECT,
    list: DataType.ARRAY,
}


def _python_args_to_sproc_args(parameters):
    """
    Converts a dictionary of Python arguments to a list of Snowflake stored procedure arguments.
    """
    sproc_args = []
    for name, param in parameters.items():
        if name == "sp_session":
            continue
        arg = {"name": name, "data_type": _PYTHON_TO_SNOWFLAKE_TYPE[param.annotation]}
        if param.default is not inspect.Parameter.empty:
            arg["default"] = param.default
        sproc_args.append(arg)
    return sproc_args


def procedure(schema="PUBLIC"):
    def decorator(func):
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        wrapper.is_procedure = True
        wrapper.sproc_args = _python_args_to_sproc_args(inspect.signature(func).parameters)
        wrapper.schema = schema
        return wrapper

    return decorator


def install(sp_session):
    """
    Installs the titan spi functions and procedures into the current database.
    """

    conn = sp_session.connection

    session_ctx = dp.fetch_session(conn)
    if "SYSADMIN" != session_ctx["role"]:
        # TODO: make dynamic role selection work
        raise Exception(f"You must use the SYSADMIN role to install the Titan SPI [role={session_ctx['role']}]")

    visible_stages = dp.list_stages(conn)
    stage = None
    for s in visible_stages:
        if s["url"] == "s3://titan-snowflake/":
            stage = s
            break

    if stage is None:
        raise Exception(
            "Cannot find Titan stage. Did you forget to run `CREATE STAGE titan_aws URL = 's3://titan-snowflake/'`?"
        )

    titan_db = stage["database_name"]

    blueprint = Blueprint("titan", database=titan_db, allow_role_switching=False)

    sprocs = []
    for name, func in vars(__this__).items():
        if callable(func) and getattr(func, "is_procedure", False):
            sprocs.append(
                resources.PythonStoredProcedure(
                    name=name,
                    owner="SYSADMIN",
                    args=func.sproc_args,
                    returns=DataType.OBJECT,
                    runtime_version="3.9",
                    packages=["snowflake-snowpark-python", "inflection", "pyparsing"],
                    imports=[f"@{stage['fqn']}/releases/titan-{__version__}.zip"],
                    handler=f"titan.spi.{name}",
                    execute_as="CALLER",
                    schema=func.schema,
                ),
            )

    blueprint.add(
        resources.Schema(name="BLUEPRINT", comment="Blueprint functions"),
        *sprocs,
    )
    plan = blueprint.plan(conn)
    result = blueprint.apply(conn, plan)
    return _to_object(
        {
            "plan": plan,
            "actions": result,
        }
    )


def _execute(sf_session, sql: list):
    for sql_text in sql:
        try:
            sf_session.execute_string(sql_text)
        except SnowparkSQLException as err:
            raise SnowparkSQLException(f"failed to execute sql, [{sql_text}]", error_code=err.error_code) from err


def _to_object(obj):
    if isinstance(obj, dict):
        return {_to_object(k): _to_object(v) for k, v in obj.items()}
    if isinstance(obj, (list, tuple)):
        return [_to_object(v) for v in obj]
    if isinstance(obj, (Action, URN)):
        return str(obj)
    if obj is None:
        return obj
    if not isinstance(obj, (int, float, str, bool)):
        raise Exception(f"Cannot convert {type(obj)} to JSON")
    return obj


###############################################################################
# Blueprint functions
###############################################################################


# @procedure()(schema="BLUEPRINT")
# def blueprint_plan(sp_session, blueprint: dict):
#     conn = sp_session.connection
#     bp = Blueprint(**blueprint)
#     plan = blueprint.plan(conn)
#     return _to_object(
#         {
#             "plan": plan,
#         }
#     )


###############################################################################
# Create or Update functions
###############################################################################


def _create_or_update_resource(
    sf_session,
    resource_type: ResourceType,
    config: dict,
    dry_run: bool = False,
):
    fqn = parse_FQN(config["name"], is_db_scoped=(resource_type == ResourceType.SCHEMA))

    resource_cls = resources.Resource.resolve_resource_cls(resource_type)
    session_ctx = dp.fetch_session(sf_session)
    if isinstance(resource_cls.scope, (DatabaseScope, SchemaScope)) and fqn.database is None:
        fqn.database = config.get("database", session_ctx["database"])
    urn = URN(resource_type=resource_type, fqn=fqn, account_locator=session_ctx["account_locator"])
    original = dp.fetch_resource(sf_session, urn)
    sql = []
    if original:
        resource_cls = resources.Resource.resolve_resource_cls(resource_type)
        original = {str(urn): dp.remove_none_values(original)}
        new = {str(urn): dp.remove_none_values(resource_cls.defaults() | config)}
        for resource_diff in diff(original, new):
            # TODO: replication groups need both sides of the diff (old and new values).
            # There are most likely other resources that need this as well.
            sql.append(lifecycle.update_resource(urn, resource_diff.new_value, resource_cls.props))
    else:
        sql = [lifecycle.create_resource(urn, config, if_not_exists=True)]
    if not dry_run:
        _execute(sf_session, sql)
    return {"sql": sql}


@procedure()
def create_or_update_database(sp_session, config: dict, dry_run: bool = False):
    return _create_or_update_resource(sp_session.connection, ResourceType.DATABASE, config, dry_run)


@procedure()
def create_or_update_schema(sp_session, config: dict, dry_run: bool = False):
    """
    Takes configuration (either as an OBJECT or a YAML string) and creates or updates a schema.
    Use the `dry_run` parameter to test the operation without executing any SQL.

    Parameters
    ----------
    config : OBJECT
        A dictionary containing the schema configuration
    dry_run : BOOLEAN
        If True, do not execute any SQL

    Schema Config
    -------------
    name : STRING
        The name of the schema
    database : STRING
        The name of the database in which to create the schema
    comment : STRING
        A comment to attach to the schema
    data_retention_time_in_days : INTEGER
        The number of days to retain data in the schema
    default_ddl_collation : STRING
        The default collation for the schema
    max_data_extension_time_in_days : INTEGER
        The maximum number of days to extend data in the schema
    owner : STRING
        The name of the user or role that owns the schema
    transient : BOOLEAN
        If True, the schema is transient
    managed_access : BOOLEAN
        If True, the schema is managed
    """
    return _create_or_update_resource(sp_session.connection, ResourceType.SCHEMA, config, dry_run)


@procedure()
def create_or_update_user(sp_session, config: dict, dry_run: bool = False):
    """
    Takes configuration (either as an OBJECT or a YAML string) and creates or updates a user.
    Use the `dry_run` parameter to test the operation without executing any SQL.

    Parameters
    ----------
    config : OBJECT
        A dictionary containing the schema configuration
    dry_run : BOOLEAN
        If True, do not execute any SQL

    User Config
    -----------
    name : STRING
        The name of the user
    comment : STRING
        A comment to attach to the user

    """
    return _create_or_update_resource(sp_session.connection, ResourceType.USER, config, dry_run)


@procedure()
def create_or_update_warehouse(sp_session, config: dict, dry_run: bool = False):
    return _create_or_update_resource(sp_session.connection, ResourceType.WAREHOUSE, config, dry_run)


@procedure()
def create_or_update_role(sp_session, config: dict, dry_run: bool = False):
    return _create_or_update_resource(sp_session.connection, ResourceType.ROLE, config, dry_run)


###############################################################################
# Fetch functions
###############################################################################


@procedure()
def fetch_database(sp_session, name: str) -> dict:
    """
    Returns a database's configuration.
    """
    return dp.fetch_database(sp_session.connection, FQN(name))


@procedure()
def fetch_schema(sp_session, name: str) -> dict:
    """
    Returns a schema's configuration.
    """
    fqn = parse_FQN(name, is_db_scoped=True)
    if fqn.database is None:
        fqn.database = sp_session.get_current_database()
    return dp.fetch_schema(sp_session.connection, fqn)


@procedure()
def fetch_user(sp_session, name: str) -> dict:
    """
    Returns a user's configuration.
    """
    return dp.fetch_user(sp_session.connection, FQN(name))


@procedure()
def fetch_warehouse(sp_session, name: str) -> dict:
    """
    Returns a user's configuration.
    """
    return dp.fetch_warehouse(sp_session.connection, FQN(name))


@procedure()
def fetch_role(sp_session, name: str) -> dict:
    """
    Returns a user's configuration.
    """
    return dp.fetch_role(sp_session.connection, FQN(name))


# def git_export(sp_session, locator: str, repo: str, path: str) -> dict:
#     access_token = _snowflake.get_generic_secret_string("github_access_token")
#     return git.export(
#         sp_session.connection,
#         repo=repo,  # Can this be configured?
#         path=path,
#         locator_str="database:TITAN",
#         access_token=access_token,
#     )


def help(_):
    txt = re.sub("[^\b]\b", "", pydoc.render_doc(__this__, "Help on %s"))
    txt = txt.split("\n\nNAME\n    ")[1].split("\n\nFILE\n")[0]
    return txt



import logging
import pytest

from titan import resources as res
from titan import Resource
from titan.enums import ResourceType
from tests.helpers import get_json_fixture, camelcase_to_snakecase


logger = logging.getLogger("titan")


def test_internal_stage():
    data = get_json_fixture("internal_stage")
    data["resource_type"] = ResourceType.STAGE
    resource = Resource.from_dict(data)
    assert isinstance(resource, res.InternalStage)


def test_external_stage():
    data = get_json_fixture("external_stage")
    data["resource_type"] = ResourceType.STAGE
    resource = Resource.from_dict(data)
    assert isinstance(resource, res.ExternalStage)


def test_table_stream():
    data = get_json_fixture("table_stream")
    data["resource_type"] = ResourceType.STREAM
    resource = Resource.from_dict(data)
    assert isinstance(resource, res.TableStream)


@pytest.mark.skip("external table streams need work")
def test_external_table_stream():
    data = get_json_fixture("external_table_stream")
    data["resource_type"] = ResourceType.STREAM
    resource = Resource.from_dict(data)
    assert isinstance(resource, res.ExternalTableStream)


def test_stage_stream():
    data = get_json_fixture("stage_stream")
    data["resource_type"] = ResourceType.STREAM
    resource = Resource.from_dict(data)
    assert isinstance(resource, res.StageStream)


def test_view_stream():
    data = get_json_fixture("view_stream")
    data["resource_type"] = ResourceType.STREAM
    resource = Resource.from_dict(data)
    assert isinstance(resource, res.ViewStream)


def enumerate_polymorphic_resources():
    resources = []
    for resource_type, class_list in Resource.__types__.items():
        if len(class_list) > 1:
            for class_ in class_list:
                resources.append((resource_type, class_))
    return resources


POLYMORPHIC_RESOURCES = enumerate_polymorphic_resources()


@pytest.fixture(
    params=POLYMORPHIC_RESOURCES,
    ids=[f"{resource_type}:{class_.__name__}" for resource_type, class_ in POLYMORPHIC_RESOURCES],
    scope="function",
)
def polymorphic_resource(request):
    resource_type, class_list = request.param
    yield resource_type, class_list


@pytest.mark.skip(reason="not a huge priority")
def test_polymorphic_resources(polymorphic_resource):
    resource_type, class_ = polymorphic_resource

    resource_name = camelcase_to_snakecase(class_.__name__)
    try:
        data = get_json_fixture(resource_name)
    except FileNotFoundError:
        pytest.fail(f"No JSON fixture for {resource_name}")
    except ValueError:
        pytest.fail(f"Missing or malformed JSON fixture for {resource_name}")
    assert Resource.resolve_resource_cls(resource_type, data) == class_, f"{resource_name} -> {class_.__name__}"



from dataclasses import dataclass, field

from ..enums import ParseableEnum, ResourceType
from ..props import BoolProp, FlagProp, IdentifierProp, Props, StringProp, TimeTravelProp
from ..resource_name import ResourceName
from ..scope import SchemaScope
from .resource import NamedResource, Resource, ResourcePointer, ResourceSpec

from .role import Role
from .table import Table
from .view import View


class StreamType(ParseableEnum):
    TABLE = "TABLE"
    EXTERNAL_TABLE = "EXTERNAL TABLE"
    STAGE = "STAGE"
    VIEW = "VIEW"


@dataclass(unsafe_hash=True)
class _TableStream(ResourceSpec):
    name: ResourceName
    on_table: Table
    owner: Role = "SYSADMIN"
    copy_grants: bool = field(default_factory=None, metadata={"fetchable": False})
    at: dict[str, str] = field(default_factory=None, metadata={"fetchable": False})
    before: dict[str, str] = field(default_factory=None, metadata={"fetchable": False})
    append_only: bool = False
    show_initial_rows: bool = field(default_factory=None, metadata={"fetchable": False})
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.at:
            self.at = {k.lower(): v for k, v in self.at.items()}
        if self.before:
            self.before = {k.lower(): v for k, v in self.before.items()}


class TableStream(NamedResource, Resource):
    """
    Description:
        Represents a stream on a table in Snowflake, which allows for change data capture on the table.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-stream

    Fields:
        name (string, required): The name of the stream.
        on_table (string, required): The name of the table the stream is based on.
        owner (string or Role): The role that owns the stream. Defaults to "SYSADMIN".
        copy_grants (bool): Whether to copy grants from the source table to the stream.
        at (dict): A dictionary specifying the point in time for the stream to start, using keys like TIMESTAMP, OFFSET, STATEMENT, or STREAM.
        before (dict): A dictionary specifying the point in time for the stream to start, similar to 'at' but defining a point before the specified time.
        append_only (bool): If set to True, the stream records only append operations.
        show_initial_rows (bool): If set to True, the stream includes the initial rows of the table at the time of stream creation.
        comment (string): An optional description for the stream.

    Python:

        ```python
        stream = TableStream(
            name="some_stream",
            on_table="some_table",
            owner="SYSADMIN",
            copy_grants=True,
            at={"TIMESTAMP": "2022-01-01 00:00:00"},
            before={"STREAM": "some_other_stream"},
            append_only=False,
            show_initial_rows=True,
            comment="This is a sample stream."
        )
        ```

    Yaml:

        ```yaml
        streams:
          - name: some_stream
            on_table: some_table
            owner: SYSADMIN
            copy_grants: true
            at:
              TIMESTAMP: "2022-01-01 00:00:00"
            before:
              STREAM: some_other_stream
            append_only: false
            show_initial_rows: true
            comment: This is a sample stream.
        ```
    """

    resource_type = ResourceType.STREAM
    props = Props(
        copy_grants=FlagProp("copy grants"),
        on_table=IdentifierProp("on table", eq=False),
        at=TimeTravelProp("at"),
        before=TimeTravelProp("before"),
        append_only=BoolProp("append_only"),
        show_initial_rows=BoolProp("show_initial_rows"),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _TableStream

    def __init__(
        self,
        name: str,
        on_table: str,
        owner: str = "SYSADMIN",
        copy_grants: bool = None,
        at: dict[str, str] = None,
        before: dict[str, str] = None,
        append_only: bool = False,
        show_initial_rows: bool = None,
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _TableStream = _TableStream(
            name=self._name,
            on_table=on_table,
            owner=owner,
            copy_grants=copy_grants,
            at=at,
            before=before,
            append_only=append_only,
            show_initial_rows=show_initial_rows,
            comment=comment,
        )
        self.requires(self._data.on_table)
        if self._data.at and "stream" in self._data.at:
            self.requires(ResourcePointer(name=self._data.at["stream"], resource_type=ResourceType.STREAM))
        if self._data.before and "stream" in self._data.before:
            self.requires(ResourcePointer(name=self._data.before["stream"], resource_type=ResourceType.STREAM))


# @dataclass(unsafe_hash=True)
# class _ExternalTableStream(ResourceSpec):
#     name: str
#     on_external_table: str
#     owner: str = "SYSADMIN"
#     copy_grants: bool = None
#     at: dict[str, str] = None
#     before: dict[str, str] = None
#     insert_only: bool = None
#     comment: str = None


# class ExternalTableStream(Resource):
#     """
#     CREATE [ OR REPLACE ] STREAM [IF NOT EXISTS]
#       <name>
#       [ COPY GRANTS ]
#       ON EXTERNAL TABLE <external_table_name>
#       [ { AT | BEFORE } ( { TIMESTAMP => <timestamp> | OFFSET => <time_difference> | STATEMENT => <id> | STREAM => '<name>' } ) ]
#       [ INSERT_ONLY = TRUE ]
#       [ COMMENT = '<string_literal>' ]
#     """

#     resource_type = ResourceType.STREAM
#     props = Props(
#         copy_grants=FlagProp("copy grants"),
#         on_external_table=IdentifierProp("on external table", eq=False),
#         at=TimeTravelProp("at"),
#         before=TimeTravelProp("before"),
#         insert_only=BoolProp("insert_only"),
#         comment=StringProp("comment"),
#     )
#     scope = SchemaScope()
#     spec = _ExternalTableStream

#     def __init__(
#         self,
#         name: str,
#         on_external_table: str,
#         owner: str = "SYSADMIN",
#         copy_grants: bool = None,
#         at: dict[str, str] = None,
#         before: dict[str, str] = None,
#         insert_only: bool = None,
#         comment: str = None,
#         **kwargs,
#     ):
#         super().__init__(**kwargs)
#         self._data = _ExternalTableStream(
#             name=name,
#             on_external_table=on_external_table,
#             owner=owner,
#             copy_grants=copy_grants,
#             at=at,
#             before=before,
#             insert_only=insert_only,
#             comment=comment,
#         )


@dataclass(unsafe_hash=True)
class _StageStream(ResourceSpec):
    name: ResourceName
    on_stage: str
    owner: Role = "SYSADMIN"
    copy_grants: bool = field(default_factory=None, metadata={"fetchable": False})
    comment: str = None


class StageStream(NamedResource, Resource):
    """
    Description:
        Represents a stream on a stage in Snowflake, which allows for capturing data changes on the stage.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-stream

    Fields:
        name (string, required): The name of the stream.
        on_stage (string, required): The name of the stage the stream is based on.
        owner (string or Role): The role that owns the stream. Defaults to "SYSADMIN".
        copy_grants (bool): Whether to copy grants from the source stage to the stream.
        comment (string): An optional description for the stream.

    Python:

        ```python
        stream = StageStream(
            name="some_stream",
            on_stage="some_stage",
            owner="SYSADMIN",
            copy_grants=True,
            comment="This is a sample stream."
        )
        ```

    Yaml:

        ```yaml
        streams:
          - name: some_stream
            on_stage: some_stage
            owner: SYSADMIN
            copy_grants: true
            comment: This is a sample stream.
        ```
    """

    resource_type = ResourceType.STREAM
    props = Props(
        copy_grants=FlagProp("copy grants"),
        on_stage=IdentifierProp("on stage", eq=False),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _StageStream

    def __init__(
        self,
        name: str,
        on_stage: str,
        owner: str = "SYSADMIN",
        copy_grants: bool = None,
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _StageStream = _StageStream(
            name=self._name,
            on_stage=on_stage,
            owner=owner,
            copy_grants=copy_grants,
            comment=comment,
        )
        self.requires(ResourcePointer(name=self._data.on_stage, resource_type=ResourceType.STAGE))


@dataclass(unsafe_hash=True)
class _ViewStream(ResourceSpec):
    name: ResourceName
    on_view: View
    owner: Role = "SYSADMIN"
    copy_grants: bool = field(default_factory=None, metadata={"fetchable": False})
    at: dict[str, str] = field(default_factory=None, metadata={"fetchable": False})
    before: dict[str, str] = field(default_factory=None, metadata={"fetchable": False})
    append_only: bool = False
    show_initial_rows: bool = field(default_factory=None, metadata={"fetchable": False})
    comment: str = None


class ViewStream(NamedResource, Resource):
    """
    Description:
        Represents a stream on a view in Snowflake, allowing for real-time data processing and querying.
        This stream can be configured with various options such as time travel, append-only mode, and initial row visibility.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-stream

    Fields:
        name (string, required): The name of the stream.
        on_view (string, required): The name of the view the stream is based on.
        owner (string or Role): The role that owns the stream. Defaults to 'SYSADMIN'.
        copy_grants (bool): Whether to copy grants from the view to the stream.
        at (dict): A dictionary specifying the point in time for the stream to start, using keys like TIMESTAMP, OFFSET, STATEMENT, or STREAM.
        before (dict): A dictionary specifying the point in time for the stream to start, similar to 'at' but defining a point before the specified time.
        append_only (bool): If set to True, the stream records only append operations.
        show_initial_rows (bool): If set to True, the stream includes the initial rows of the view at the time of stream creation.
        comment (string): An optional description for the stream.

    Python:

        ```python
        view_stream = ViewStream(
            name="some_stream",
            on_view="some_view",
            owner="SYSADMIN",
            copy_grants=True,
            at={"TIMESTAMP": "2022-01-01 00:00:00"},
            before={"STREAM": "some_other_stream"},
            append_only=False,
            show_initial_rows=True,
            comment="This is a sample stream on a view."
        )
        ```

    Yaml:

        ```yaml
        streams:
          - name: some_stream
            on_view: some_view
            owner: SYSADMIN
            copy_grants: true
            at:
              TIMESTAMP: "2022-01-01 00:00:00"
            before:
              STREAM: some_other_stream
            append_only: false
            show_initial_rows: true
            comment: This is a sample stream on a view.
        ```
    """

    resource_type = ResourceType.STREAM
    props = Props(
        copy_grants=FlagProp("copy grants"),
        on_view=IdentifierProp("on view", eq=False),
        at=TimeTravelProp("at"),
        before=TimeTravelProp("before"),
        append_only=BoolProp("append_only"),
        show_initial_rows=BoolProp("show_initial_rows"),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _ViewStream

    def __init__(
        self,
        name: str,
        on_view: str,
        owner: str = "SYSADMIN",
        copy_grants: bool = None,
        at: dict[str, str] = None,
        before: dict[str, str] = None,
        append_only: bool = None,
        show_initial_rows: bool = None,
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _ViewStream = _ViewStream(
            name=self._name,
            on_view=on_view,
            owner=owner,
            copy_grants=copy_grants,
            at=at,
            before=before,
            append_only=append_only,
            show_initial_rows=show_initial_rows,
            comment=comment,
        )
        self.requires(self._data.on_view)


StreamTypeMap = {
    StreamType.TABLE: TableStream,
    # StreamType.EXTERNAL_TABLE: ExternalTableStream,
    StreamType.STAGE: StageStream,
    StreamType.VIEW: ViewStream,
}


def _resolver(data: dict):
    if "on_table" in data:
        return TableStream
    # elif "on_external_table" in data:
    #     return ExternalTableStream
    elif "on_stage" in data:
        return StageStream
    elif "on_view" in data:
        return ViewStream
    # using this as a workaround because there may not be enough properties during a small change to disambiguate
    # really the different stream types should probably have seperate resource types.
    # Either that, or the resolver would need to look at the database to see what type of stream it is
    return TableStream


Resource.__resolvers__[ResourceType.STREAM] = _resolver



from dataclasses import dataclass

from ..enums import ParseableEnum, ResourceType
from ..props import EnumProp, Props, StringListProp, StringProp
from ..resource_name import ResourceName
from ..scope import SchemaScope
from .resource import NamedResource, Resource, ResourceSpec
from .role import Role


class AuthenticationMethods(ParseableEnum):
    ALL = "ALL"
    PASSWORD = "PASSWORD"
    SAML = "SAML"
    OAUTH = "OAUTH"
    KEYPAIR = "KEYPAIR"


class MFAEnrollment(ParseableEnum):
    REQUIRED = "REQUIRED"
    OPTIONAL = "OPTIONAL"


class ClientTypes(ParseableEnum):
    ALL = "ALL"
    SNOWFLAKE_UI = "SNOWFLAKE_UI"
    DRIVERS = "DRIVERS"
    SNOWSQL = "SNOWSQL"


@dataclass(unsafe_hash=True)
class _AuthenticationPolicy(ResourceSpec):
    name: ResourceName
    authentication_methods: list[AuthenticationMethods] = None
    mfa_authentication_methods: list[AuthenticationMethods] = None
    mfa_enrollment: MFAEnrollment = "OPTIONAL"
    client_types: list[ClientTypes] = None
    security_integrations: list[str] = None
    comment: str = None
    owner: Role = "SECURITYADMIN"

    def __post_init__(self):
        super().__post_init__()
        if self.authentication_methods is None:
            self.authentication_methods = [AuthenticationMethods.ALL]

        if self.mfa_authentication_methods is None:
            self.mfa_authentication_methods = [AuthenticationMethods.PASSWORD, AuthenticationMethods.SAML]
        else:
            for method in self.mfa_authentication_methods:
                if method not in (
                    AuthenticationMethods.ALL,
                    AuthenticationMethods.SAML,
                    AuthenticationMethods.PASSWORD,
                ):
                    raise ValueError("MFA authentication methods must be either 'ALL', 'SAML', or 'PASSWORD'")
            if (
                len(self.mfa_authentication_methods) == 1
                and self.mfa_authentication_methods[0] == AuthenticationMethods.ALL
            ):
                self.mfa_authentication_methods = [AuthenticationMethods.PASSWORD, AuthenticationMethods.SAML]

        if self.client_types is None:
            self.client_types = [ClientTypes.ALL]

        if self.security_integrations is None:
            self.security_integrations = ["ALL"]


class AuthenticationPolicy(NamedResource, Resource):
    """
    Description:
        Defines the rules and constraints for authentication within the system, ensuring they meet specific security standards.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-authentication-policy

    Fields:
        name (string, required): The name of the authentication policy.
        authentication_methods (list): A list of allowed authentication methods.
        mfa_authentication_methods (list): A list of authentication methods that enforce multi-factor authentication (MFA).
        mfa_enrollment (string): Determines whether a user must enroll in multi-factor authentication. Defaults to OPTIONAL.
        client_types (list): A list of clients that can authenticate with Snowflake.
        security_integrations (list): A list of security integrations the authentication policy is associated with.
        comment (string): A comment or description for the authentication policy.
        owner (string or Role): The owner role of the authentication policy. Defaults to SECURITYADMIN.

    Python:

        ```python
        authentication_policy = AuthenticationPolicy(
            name="some_authentication_policy",
            authentication_methods=["PASSWORD", "SAML"],
            mfa_authentication_methods=["PASSWORD"],
            mfa_enrollment="REQUIRED",
            client_types=["SNOWFLAKE_UI"],
            security_integrations=["ALL"],
            comment="Policy for secure authentication."
        )
        ```

    Yaml:

        ```yaml
        authentication_policies:
          - name: some_authentication_policy
            authentication_methods:
              - PASSWORD
              - SAML
            mfa_authentication_methods:
              - PASSWORD
            mfa_enrollment: REQUIRED
            client_types:
              - SNOWFLAKE_UI
            security_integrations:
              - ALL
            comment: Policy for secure authentication.
        ```
    """

    resource_type = ResourceType.AUTHENTICATION_POLICY
    props = Props(
        authentication_methods=StringListProp("authentication_methods", parens=True),
        mfa_authentication_methods=StringListProp("mfa_authentication_methods", parens=True),
        mfa_enrollment=EnumProp("mfa_enrollment", MFAEnrollment),
        client_types=StringListProp("client_types", parens=True),
        security_integrations=StringListProp("security_integrations", parens=True),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _AuthenticationPolicy

    def __init__(
        self,
        name: str,
        authentication_methods: list[str] = None,
        mfa_authentication_methods: list[str] = None,
        mfa_enrollment: str = "OPTIONAL",
        client_types: list[str] = None,
        security_integrations: list[str] = None,
        comment: str = None,
        owner: str = "SECURITYADMIN",
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _AuthenticationPolicy = _AuthenticationPolicy(
            name=self._name,
            authentication_methods=authentication_methods,
            mfa_authentication_methods=mfa_authentication_methods,
            mfa_enrollment=mfa_enrollment,
            client_types=client_types,
            security_integrations=security_integrations,
            comment=comment,
            owner=owner,
        )



"""
Policies are used to enforce constraints over a Titan project.

All Titan projects use the Titan Standard Policy by default
"""

from enum import Enum
from typing import Callable


class EnforcementLevel(Enum):
    ADVISORY = "ADVISORY"
    MANDATORY = "MANDATORY"


class Policy:
    def __init__(self, name: str, description: str, enforcement_level: str, validate: Callable):
        self.name = name
        self.description = description
        self.enforcement_level = enforcement_level
        self.validate = validate


class OwnershipPolicy(Policy):
    pass


class PolicyPack:
    def __init__(self, name: str, policies):
        self.name = name
        self.policies = policies



import pytest

from titan.parse import format_collection_string, parse_collection_string


# Test cases for parse_collection_string
def test_parse_database_level():
    assert parse_collection_string("SOME_DATABASE.<TABLE>") == {
        "in_name": "SOME_DATABASE",
        "in_type": "database",
        "on_type": "TABLE",
    }


def test_parse_schema_level():
    assert parse_collection_string("SOME_DATABASE.SOME_SCHEMA.<VIEW>") == {
        "in_name": "SOME_DATABASE.SOME_SCHEMA",
        "in_type": "schema",
        "on_type": "VIEW",
    }


def test_parse_invalid_format():
    with pytest.raises(ValueError):
        parse_collection_string("SOME_DATABASE")


def test_parse_incorrect_brackets():
    with pytest.raises(ValueError):
        parse_collection_string("SOME_DATABASE.<TABLE")


# Test cases for format_collection_string
def test_format_database_level():
    collection_dict = {"in_name": "SOME_DATABASE", "in_type": "database", "on_type": "TABLE"}
    assert format_collection_string(collection_dict) == "SOME_DATABASE.<TABLE>"


def test_format_schema_level():
    collection_dict = {"in_name": "SOME_DATABASE.SOME_SCHEMA", "in_type": "schema", "on_type": "VIEW"}
    assert format_collection_string(collection_dict) == "SOME_DATABASE.SOME_SCHEMA.<VIEW>"



from dataclasses import dataclass

from ..builtins import SYSTEM_SCHEMAS
from ..enums import ResourceType
from ..props import FlagProp, IntProp, Props, StringProp, TagsProp
from ..resource_name import ResourceName
from ..scope import DatabaseScope
from .resource import NamedResource, Resource, ResourceContainer, ResourceSpec
from .role import Role
from .tag import TaggableResource


@dataclass(unsafe_hash=True)
class _Schema(ResourceSpec):
    name: ResourceName
    transient: bool = False
    managed_access: bool = False
    data_retention_time_in_days: int = 1
    max_data_extension_time_in_days: int = 14
    default_ddl_collation: str = None
    owner: Role = "SYSADMIN"
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.transient and self.data_retention_time_in_days is not None:
            raise ValueError("Transient schema can't have data retention time")
        elif not self.transient and self.data_retention_time_in_days is None:
            self.data_retention_time_in_days = 1


class Schema(NamedResource, TaggableResource, Resource, ResourceContainer):
    """
    Description:
        Represents a schema in Snowflake, which is a logical grouping of database objects such as tables, views, and stored procedures. Schemas are used to organize and manage such objects within a database.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-schema

    Fields:
        name (string, required): The name of the schema.
        transient (bool): Specifies if the schema is transient. Defaults to False.
        managed_access (bool): Specifies if the schema has managed access. Defaults to False.
        data_retention_time_in_days (int): The number of days to retain data. Defaults to 1.
        max_data_extension_time_in_days (int): The maximum number of days to extend data retention. Defaults to 14.
        default_ddl_collation (string): The default DDL collation setting.
        tags (dict): Tags associated with the schema.
        owner (string or Role): The owner of the schema. Defaults to "SYSADMIN".
        comment (string): A comment about the schema.

    Python:

        ```python
        schema = Schema(
            name="some_schema",
            transient=True,
            managed_access=True,
            data_retention_time_in_days=7,
            max_data_extension_time_in_days=28,
            default_ddl_collation="utf8",
            tags={"project": "analytics"},
            owner="SYSADMIN",
            comment="Schema for analytics project."
        )
        ```

    Yaml:

        ```yaml
        schemas:
          - name: some_schema
            transient: true
            managed_access: true
            data_retention_time_in_days: 7
            max_data_extension_time_in_days: 28
            default_ddl_collation: utf8
            tags:
              project: analytics
            owner: SYSADMIN
            comment: Schema for analytics project.
        ```
    """

    resource_type = ResourceType.SCHEMA
    props = Props(
        transient=FlagProp("transient"),
        managed_access=FlagProp("with managed access"),
        data_retention_time_in_days=IntProp("data_retention_time_in_days"),
        max_data_extension_time_in_days=IntProp("max_data_extension_time_in_days"),
        default_ddl_collation=StringProp("default_ddl_collation"),
        tags=TagsProp(),
        comment=StringProp("comment"),
    )
    scope = DatabaseScope()
    spec = _Schema

    def __init__(
        self,
        name: str,
        transient: bool = False,
        managed_access: bool = False,
        data_retention_time_in_days: int = None,
        max_data_extension_time_in_days: int = 14,
        default_ddl_collation: str = None,
        tags: dict[str, str] = None,
        owner: str = "SYSADMIN",
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)

        if self._name == "INFORMATION_SCHEMA":
            comment = "Views describing the contents of schemas in this database"
            owner = ""

        # if self._name in SYSTEM_SCHEMAS:
        #     self.implicit = True

        self._data: _Schema = _Schema(
            name=self._name,
            transient=transient,
            managed_access=managed_access,
            data_retention_time_in_days=data_retention_time_in_days,
            max_data_extension_time_in_days=max_data_extension_time_in_days,
            default_ddl_collation=default_ddl_collation,
            owner=owner,
            comment=comment,
        )
        self.set_tags(tags)



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import ResourceType, Language
from ..scope import SchemaScope
from ..resource_name import ResourceName
from ..props import (
    EnumProp,
    Props,
    StringListProp,
    StringProp,
)


@dataclass(unsafe_hash=True)
class _PackagesPolicy(ResourceSpec):
    name: ResourceName
    language: Language = Language.PYTHON
    allowlist: list[str] = None
    blocklist: list[str] = None
    additional_creation_blocklist: list[str] = None
    comment: str = None
    owner: Role = "SYSADMIN"

    def __post_init__(self):
        super().__post_init__()
        if self.language != Language.PYTHON:
            raise ValueError("Language must be PYTHON")


class PackagesPolicy(NamedResource, Resource):
    """
    Description:
        A Packages Policy defines a set of rules for allowed and blocked packages
        that are applied to user-defined functions and stored procedures.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-packages-policy

    Fields:
        name (string, required): The name of the packages policy.
        language (string or Language): The programming language for the packages. Defaults to PYTHON.
        allowlist (list): A list of package specifications that are explicitly allowed.
        blocklist (list): A list of package specifications that are explicitly blocked.
        additional_creation_blocklist (list): A list of package specifications that are blocked during creation.
        comment (string): A comment or description for the packages policy.
        owner (string or Role): The owner role of the packages policy. Defaults to SYSADMIN.

    Python:

        ```python
        packages_policy = PackagesPolicy(
            name="some_packages_policy",
            allowlist=["numpy", "pandas"],
            blocklist=["os", "sys"],
            comment="Policy for data processing packages."
        )
        ```

    Yaml:

        ```yaml
        packages_policy:
          - name: some_packages_policy
            allowlist:
              - numpy
              - pandas
            blocklist:
              - os
              - sys
            comment: Policy for data processing packages.
        ```
    """

    resource_type = ResourceType.PACKAGES_POLICY
    props = Props(
        language=EnumProp("language", [Language.PYTHON], eq=False),
        allowlist=StringListProp("allowlist", parens=True),
        blocklist=StringListProp("blocklist", parens=True),
        additional_creation_blocklist=StringListProp("additional_creation_blocklist", parens=True),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _PackagesPolicy

    def __init__(
        self,
        name: str,
        language: Language = Language.PYTHON,
        allowlist: list[str] = None,
        blocklist: list[str] = None,
        additional_creation_blocklist: list[str] = None,
        comment: str = None,
        owner: str = "SYSADMIN",
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _PackagesPolicy = _PackagesPolicy(
            name=self._name,
            language=language,
            allowlist=allowlist,
            blocklist=blocklist,
            additional_creation_blocklist=additional_creation_blocklist,
            comment=comment,
            owner=owner,
        )



from dataclasses import dataclass


from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import ParseableEnum, ResourceType
from ..props import Props, StringProp, BoolProp, EnumProp, StringListProp
from ..scope import AccountScope
from ..resource_name import ResourceName


class NotificationType(ParseableEnum):
    EMAIL = "EMAIL"
    QUEUE = "QUEUE"


class NotificationProvider(ParseableEnum):
    AWS_SNS = "AWS_SNS"
    GCP_PUBSUB = "GCP_PUBSUB"
    AZURE_STORAGE_QUEUE = "AZURE_STORAGE_QUEUE"
    AZURE_EVENT_GRID = "AZURE_EVENT_GRID"


class NotificationDirection(ParseableEnum):
    INBOUND = "INBOUND"
    OUTBOUND = "OUTBOUND"


@dataclass(unsafe_hash=True)
class _EmailNotificationIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    allowed_recipients: list[str]
    type: NotificationType = NotificationType.EMAIL
    owner: Role = "ACCOUNTADMIN"
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.allowed_recipients is not None and len(self.allowed_recipients) == 0:
            raise ValueError("allowed_recipients can't be empty")


class EmailNotificationIntegration(NamedResource, Resource):
    """
    Description:
        Manages the configuration for email-based notification integrations within Snowflake. This integration
        allows specifying recipients who will receive notifications via email.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-notification-integration

    Fields:
        name (string, required): The name of the email notification integration.
        enabled (bool, required): Specifies whether the notification integration is enabled.
        allowed_recipients (list): A list of email addresses that are allowed to receive notifications.
        comment (string): An optional comment about the notification integration.
        owner (string or Role): The owner role of the notification integration. Defaults to "ACCOUNTADMIN".

    Python:

        ```python
        email_notification_integration = EmailNotificationIntegration(
            name="some_email_notification_integration",
            enabled=True,
            allowed_recipients=["user1@example.com", "user2@example.com"],
            comment="Example email notification integration"
        )
        ```

    Yaml:

        ```yaml
        email_notification_integrations:
          - name: some_email_notification_integration
            enabled: true
            allowed_recipients:
              - user1@example.com
              - user2@example.com
            comment: "Example email notification integration"
        ```
    """

    resource_type = ResourceType.NOTIFICATION_INTEGRATION
    props = Props(
        type=EnumProp("type", [NotificationType.EMAIL]),
        enabled=BoolProp("enabled"),
        allowed_recipients=StringListProp("allowed_recipients", parens=True),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _EmailNotificationIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        allowed_recipients: list[str],
        comment: str = None,
        owner: str = "ACCOUNTADMIN",
        **kwargs,
    ):
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        self._data: _EmailNotificationIntegration = _EmailNotificationIntegration(
            name=self._name,
            enabled=enabled,
            allowed_recipients=allowed_recipients,
            comment=comment,
            owner=owner,
        )


@dataclass(unsafe_hash=True)
class _AWSOutboundNotificationIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    aws_sns_topic_arn: str
    aws_sns_role_arn: str
    direction: NotificationDirection = NotificationDirection.OUTBOUND
    type: str = NotificationType.QUEUE
    notification_provider: str = NotificationProvider.AWS_SNS
    owner: Role = "ACCOUNTADMIN"
    comment: str = None


class AWSOutboundNotificationIntegration(NamedResource, Resource):
    """
    CREATE [ OR REPLACE ] NOTIFICATION INTEGRATION [IF NOT EXISTS]
      <name>
      ENABLED = { TRUE | FALSE }
      DIRECTION = OUTBOUND
      TYPE = QUEUE
      cloudProviderParamsPush
      [ COMMENT = '<string_literal>' ]

    cloudProviderParamsPush (for Amazon SNS) ::=

      NOTIFICATION_PROVIDER = AWS_SNS
      AWS_SNS_TOPIC_ARN = '<topic_arn>'
      AWS_SNS_ROLE_ARN = '<iam_role_arn>'
    """

    resource_type = ResourceType.NOTIFICATION_INTEGRATION
    props = Props(
        enabled=BoolProp("enabled"),
        direction=EnumProp("direction", [NotificationDirection.OUTBOUND]),
        type=EnumProp("type", [NotificationType.QUEUE]),
        notification_provider=EnumProp("notification_provider", [NotificationProvider.AWS_SNS]),
        aws_sns_topic_arn=StringProp("aws_sns_topic_arn"),
        aws_sns_role_arn=StringProp("aws_sns_role_arn"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _AWSOutboundNotificationIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        aws_sns_topic_arn: str,
        aws_sns_role_arn: str,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        kwargs.pop("direction", None)
        kwargs.pop("notification_provider", None)
        super().__init__(name, **kwargs)
        self._data: _AWSOutboundNotificationIntegration = _AWSOutboundNotificationIntegration(
            name=self._name,
            enabled=enabled,
            aws_sns_topic_arn=aws_sns_topic_arn,
            aws_sns_role_arn=aws_sns_role_arn,
            owner=owner,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _GCPOutboundNotificationIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    gcp_pubsub_topic_name: str
    direction: NotificationDirection = NotificationDirection.OUTBOUND
    type: str = NotificationType.QUEUE
    notification_provider: str = NotificationProvider.GCP_PUBSUB
    owner: Role = "ACCOUNTADMIN"
    comment: str = None


class GCPOutboundNotificationIntegration(NamedResource, Resource):
    """
    CREATE [ OR REPLACE ] NOTIFICATION INTEGRATION [IF NOT EXISTS]
      <name>
      ENABLED = { TRUE | FALSE }
      DIRECTION = OUTBOUND
      TYPE = QUEUE
      cloudProviderParamsPush
      [ COMMENT = '<string_literal>' ]

    cloudProviderParamsPush (for Google Pub/Sub) ::=
      NOTIFICATION_PROVIDER = GCP_PUBSUB
      GCP_PUBSUB_TOPIC_NAME = '<topic_id>'
    """

    resource_type = ResourceType.NOTIFICATION_INTEGRATION
    props = Props(
        enabled=BoolProp("enabled"),
        direction=EnumProp("direction", [NotificationDirection.OUTBOUND]),
        type=EnumProp("type", [NotificationType.QUEUE]),
        notification_provider=EnumProp("notification_provider", [NotificationProvider.GCP_PUBSUB]),
        gcp_pubsub_topic_name=StringProp("gcp_pubsub_topic_name"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _GCPOutboundNotificationIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        gcp_pubsub_topic_name: str,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        kwargs.pop("direction", None)
        kwargs.pop("notification_provider", None)
        super().__init__(name, **kwargs)
        self._data: _GCPOutboundNotificationIntegration = _GCPOutboundNotificationIntegration(
            name=self._name,
            enabled=enabled,
            gcp_pubsub_topic_name=gcp_pubsub_topic_name,
            owner=owner,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _AzureOutboundNotificationIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    azure_event_grid_topic_endpoint: str
    azure_tenant_id: str
    direction: NotificationDirection = NotificationDirection.OUTBOUND
    type: str = NotificationType.QUEUE
    notification_provider: str = NotificationProvider.AZURE_EVENT_GRID
    owner: Role = "ACCOUNTADMIN"
    comment: str = None


class AzureOutboundNotificationIntegration(NamedResource, Resource):
    """
    CREATE [ OR REPLACE ] NOTIFICATION INTEGRATION [IF NOT EXISTS]
      <name>
      ENABLED = { TRUE | FALSE }
      DIRECTION = OUTBOUND
      TYPE = QUEUE
      cloudProviderParamsPush
      [ COMMENT = '<string_literal>' ]

    cloudProviderParamsPush (for Microsoft Azure Event Grid) ::=
      NOTIFICATION_PROVIDER = AZURE_EVENT_GRID
      AZURE_EVENT_GRID_TOPIC_ENDPOINT = '<event_grid_topic_endpoint>'
      AZURE_TENANT_ID = '<directory_ID>';
    """

    resource_type = ResourceType.NOTIFICATION_INTEGRATION
    props = Props(
        enabled=BoolProp("enabled"),
        direction=EnumProp("direction", [NotificationDirection.OUTBOUND]),
        type=EnumProp("type", [NotificationType.QUEUE]),
        notification_provider=EnumProp("notification_provider", [NotificationProvider.AZURE_EVENT_GRID]),
        azure_event_grid_topic_endpoint=StringProp("azure_event_grid_topic_endpoint"),
        azure_tenant_id=StringProp("azure_tenant_id"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _AzureOutboundNotificationIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        azure_event_grid_topic_endpoint: str,
        azure_tenant_id: str,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        kwargs.pop("direction", None)
        kwargs.pop("notification_provider", None)
        super().__init__(name, **kwargs)
        self._data: _AzureOutboundNotificationIntegration = _AzureOutboundNotificationIntegration(
            name=self._name,
            enabled=enabled,
            azure_event_grid_topic_endpoint=azure_event_grid_topic_endpoint,
            azure_tenant_id=azure_tenant_id,
            owner=owner,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _GCPInboundNotificationIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    gcp_pubsub_subscription_name: str
    direction: NotificationDirection = NotificationDirection.INBOUND
    type: str = NotificationType.QUEUE
    notification_provider: str = NotificationProvider.GCP_PUBSUB
    owner: Role = "ACCOUNTADMIN"
    comment: str = None


class GCPInboundNotificationIntegration(NamedResource, Resource):
    """
    CREATE [ OR REPLACE ] NOTIFICATION INTEGRATION [IF NOT EXISTS]
      <name>
      ENABLED = { TRUE | FALSE }
      TYPE = QUEUE
      cloudProviderParamsAuto
      [ COMMENT = '<string_literal>' ]

    cloudProviderParamsAuto (for Google Cloud Storage) ::=
      NOTIFICATION_PROVIDER = GCP_PUBSUB
      GCP_PUBSUB_SUBSCRIPTION_NAME = '<subscription_id>'
    """

    resource_type = ResourceType.NOTIFICATION_INTEGRATION
    props = Props(
        enabled=BoolProp("enabled"),
        type=EnumProp("type", [NotificationType.QUEUE]),
        notification_provider=EnumProp("notification_provider", [NotificationProvider.GCP_PUBSUB]),
        gcp_pubsub_subscription_name=StringProp("gcp_pubsub_subscription_name"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _GCPInboundNotificationIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        gcp_pubsub_subscription_name: str,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        kwargs.pop("notification_provider", None)
        super().__init__(name, **kwargs)
        self._data: _GCPInboundNotificationIntegration = _GCPInboundNotificationIntegration(
            name=self._name,
            enabled=enabled,
            gcp_pubsub_subscription_name=gcp_pubsub_subscription_name,
            owner=owner,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _AzureInboundNotificationIntegration(ResourceSpec):
    name: ResourceName
    enabled: bool
    azure_storage_queue_primary_uri: str
    azure_tenant_id: str
    direction: NotificationDirection = NotificationDirection.INBOUND
    type: str = NotificationType.QUEUE
    notification_provider: str = NotificationProvider.AZURE_STORAGE_QUEUE
    owner: Role = "ACCOUNTADMIN"
    comment: str = None


class AzureInboundNotificationIntegration(NamedResource, Resource):
    """
    CREATE [ OR REPLACE ] NOTIFICATION INTEGRATION [IF NOT EXISTS]
      <name>
      ENABLED = { TRUE | FALSE }
      TYPE = QUEUE
      cloudProviderParamsAuto
      [ COMMENT = '<string_literal>' ]

    cloudProviderParamsAuto (for Microsoft Azure Storage) ::=
      NOTIFICATION_PROVIDER = AZURE_STORAGE_QUEUE
      AZURE_STORAGE_QUEUE_PRIMARY_URI = '<queue_URL>'
      AZURE_TENANT_ID = '<directory_ID>';
    """

    resource_type = ResourceType.NOTIFICATION_INTEGRATION
    props = Props(
        enabled=BoolProp("enabled"),
        type=EnumProp("type", [NotificationType.QUEUE]),
        notification_provider=EnumProp("notification_provider", [NotificationProvider.AZURE_STORAGE_QUEUE]),
        azure_storage_queue_primary_uri=StringProp("azure_storage_queue_primary_uri"),
        azure_tenant_id=StringProp("azure_tenant_id"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _AzureInboundNotificationIntegration

    def __init__(
        self,
        name: str,
        enabled: bool,
        azure_storage_queue_primary_uri: str,
        azure_tenant_id: str,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        kwargs.pop("notification_provider", None)
        super().__init__(name, **kwargs)
        self._data: _AzureInboundNotificationIntegration = _AzureInboundNotificationIntegration(
            name=self._name,
            enabled=enabled,
            azure_storage_queue_primary_uri=azure_storage_queue_primary_uri,
            azure_tenant_id=azure_tenant_id,
            owner=owner,
            comment=comment,
        )


def _notification_resolver(data: dict):
    if NotificationType(data["type"]) == NotificationType.EMAIL:
        return EmailNotificationIntegration
    if "direction" in data and "notification_provider" in data:
        direction = NotificationDirection(data["direction"])
        provider = NotificationProvider(data["notification_provider"])
        if direction == NotificationDirection.INBOUND:
            if provider == NotificationProvider.AZURE_STORAGE_QUEUE:
                return AzureInboundNotificationIntegration
            elif provider == NotificationProvider.GCP_PUBSUB:
                return GCPInboundNotificationIntegration
        elif direction == NotificationDirection.OUTBOUND:
            if provider == NotificationProvider.AWS_SNS:
                return AWSOutboundNotificationIntegration
            elif provider == NotificationProvider.AZURE_EVENT_GRID:
                return AzureOutboundNotificationIntegration
            elif provider == NotificationProvider.GCP_PUBSUB:
                return GCPOutboundNotificationIntegration
    raise ValueError("Invalid direction or notification provider")


Resource.__resolvers__[ResourceType.NOTIFICATION_INTEGRATION] = _notification_resolver



from dataclasses import dataclass

from ..enums import AccountEdition, ResourceType
from ..props import Props
from ..resource_name import ResourceName
from ..scope import OrganizationScope
from .resource import NamedResource, Resource, ResourceContainer, ResourceSpec


@dataclass(unsafe_hash=True)
class _Account(ResourceSpec):
    name: ResourceName
    locator: str
    edition: AccountEdition = None
    region: str = None
    comment: str = None


class Account(NamedResource, Resource, ResourceContainer):
    """
    CREATE ACCOUNT <name>
        ADMIN_NAME = <string>
        { ADMIN_PASSWORD = '<string_literal>' | ADMIN_RSA_PUBLIC_KEY = <string> }
        [ FIRST_NAME = <string> ]
        [ LAST_NAME = <string> ]
        EMAIL = '<string>'
        [ MUST_CHANGE_PASSWORD = { TRUE | FALSE } ]
        EDITION = { STANDARD | ENTERPRISE | BUSINESS_CRITICAL }
        [ REGION_GROUP = <region_group_id> ]
        [ REGION = <snowflake_region_id> ]
        [ COMMENT = '<string_literal>' ]
    """

    resource_type = ResourceType.ACCOUNT
    props = Props()
    scope = OrganizationScope()
    spec = _Account

    def __init__(
        self,
        name: str,
        locator: str,
        edition: AccountEdition = None,
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name=name, **kwargs)
        self._data: _Account = _Account(
            name=self._name,
            locator=locator,
            edition=edition,
            comment=comment,
        )



from dataclasses import dataclass

from ..enums import ResourceType
from ..props import AlertConditionProp, Props, QueryProp, StringProp, TagsProp
from ..resource_name import ResourceName
from ..scope import SchemaScope
from .resource import NamedResource, Resource, ResourceSpec
from .role import Role
from .tag import TaggableResource
from .warehouse import Warehouse


@dataclass(unsafe_hash=True)
class _Alert(ResourceSpec):
    name: ResourceName
    warehouse: Warehouse
    schedule: str
    condition: str
    then: str
    owner: Role = "SYSADMIN"
    comment: str = None


class Alert(NamedResource, TaggableResource, Resource):
    """
    Description:
        Alerts trigger notifications when certain conditions are met.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-alert

    Fields:
        name (string, required): The name of the alert.
        warehouse (string or Warehouse): The name of the warehouse to run the query on.
        schedule (string): The schedule for the alert to run on.
        condition (string): The condition for the alert to trigger on.
        then (string): The query to run when the alert triggers.
        owner (string or Role): The owner role of the alert. Defaults to "SYSADMIN".
        comment (string): A comment for the alert. Defaults to None.
        tags (dict): Tags for the alert. Defaults to None.

    Python:

        ```python
        alert = Alert(
            name="some_alert",
            warehouse="some_warehouse",
            schedule="USING CRON * * * * *",
            condition="SELECT COUNT(*) FROM some_table",
            then="CALL SYSTEM$SEND_EMAIL('example@example.com', 'Alert Triggered', 'The alert condition was met.')",
        )
        ```

    Yaml:

        ```yaml
        alerts:
          - name: some_alert
            warehouse: some_warehouse
            schedule: USING CRON * * * * *
            condition: SELECT COUNT(*) FROM some_table
            then: CALL SYSTEM$SEND_EMAIL('example@example.com', 'Alert Triggered', 'The alert condition was met.')
        ```
    """

    resource_type = ResourceType.ALERT
    props = Props(
        warehouse=StringProp("warehouse"),
        schedule=StringProp("schedule"),
        comment=StringProp("comment"),
        tags=TagsProp(),
        condition=AlertConditionProp(),
        then=QueryProp("then"),
    )
    scope = SchemaScope()
    spec = _Alert

    def __init__(
        self,
        name: str,
        warehouse: Warehouse,
        schedule: str,
        condition: str,
        then: str,
        owner: str = "SYSADMIN",
        comment: str = None,
        tags: dict[str, str] = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _Alert = _Alert(
            name=self._name,
            warehouse=warehouse,
            schedule=schedule,
            condition=condition,
            then=then,
            owner=owner,
            comment=comment,
        )
        self.set_tags(tags)
        self.requires(self._data.warehouse)



import pytest

from titan import resources as res
from tests.helpers import get_sql_fixtures
from titan.enums import ResourceType, WarehouseSize
from titan.resource_name import ResourceName
from titan.resource_tags import ResourceTags
from titan.resources.resource import ResourcePointer

SQL_FIXTURES = list(get_sql_fixtures())


def test_resource_init_with_dict_pointer():
    res.Task(**{"name": "TASK", "as_": "SELECT 1", "warehouse": {"name": "wh"}})


def test_resource_init_with_resource_pointer():
    res.Task(name="TASK", schedule="1 minute", as_="SELECT 1", warehouse=res.Warehouse(name="wh"))


def test_resource_init_with_resource_name():
    res.Task(name="TASK", schedule="1 minute", as_="SELECT 1", warehouse="wh")


def test_resource_init_with_type():
    res.Task(**{"name": "TASK", "as_": "SELECT 1", "warehouse": {"name": "wh"}, "resource_type": "TASK"})


def test_resource_init_from_dict():
    res.Resource.from_dict({"name": "TASK", "as_": "SELECT 1", "warehouse": {"name": "wh"}, "resource_type": "TASK"})


def test_view_fails_with_empty_columns():
    with pytest.raises(ValueError):
        res.View(name="MY_VIEW", columns=[], as_="SELECT 1")


def test_view_with_columns():
    view = res.View.from_sql("CREATE VIEW MY_VIEW (COL1) AS SELECT 1")
    assert view._data.columns == [{"name": "COL1"}]


def test_enum_field_serialization():
    warehouse = res.Warehouse(name="WH", warehouse_size="XSMALL")
    assert warehouse._data.warehouse_size == WarehouseSize.XSMALL


@pytest.fixture(
    params=SQL_FIXTURES,
    ids=[f"{resource_cls.__name__}({idx})" for resource_cls, _, idx in SQL_FIXTURES],
    scope="function",
)
def sql_fixture(request):
    resource_cls, data, idx = request.param
    yield resource_cls, data


def test_init_from_sql(sql_fixture):
    resource_cls, data = sql_fixture
    try:
        resource_cls.from_sql(data)
    except Exception:
        pytest.fail(f"Failed to construct {resource_cls.__name__} from SQL fixture")


def test_resource_name_serialization():
    task = res.Task(name="TASK")
    assert task.name == "TASK"
    assert task.name == ResourceName("task")
    assert task.to_dict()["name"] == "TASK"
    assert task.fqn.name == "TASK"


def test_resource_quoted_name_serialization():
    name_str_raw = "~task"
    name_str_quoted = f'"{name_str_raw}"'
    task = res.Task(name=name_str_raw)
    assert task.name == name_str_raw
    assert task.name == ResourceName(name_str_quoted)
    assert task.to_dict()["name"] == name_str_quoted
    assert task.fqn.name == name_str_quoted


def test_resource_cased_quoted_name_serialization():
    name_str_raw = "SomeTask"
    name_str_quoted = f'"{name_str_raw}"'
    task = res.Task(name=name_str_quoted)
    assert task.name != name_str_raw
    assert task.name == name_str_quoted
    assert task.to_dict()["name"] == name_str_quoted
    assert task.fqn.name == name_str_quoted


def test_tags_definition():
    db = res.Database(name="DB", tags={"project": "test_deployment", "priority": "low"})
    assert db.tags is not None
    assert db.tags.to_dict() == {"project": "test_deployment", "priority": "low"}

    db = res.Database(name="DB", tags=ResourceTags({"project": "test_deployment", "priority": "low"}))
    assert db.tags is not None
    assert db.tags.to_dict() == {"project": "test_deployment", "priority": "low"}


def test_database_scoped_container_construction():
    db = res.Database(name="my_database")
    schema = res.Schema(name="my_schema", database=db)
    assert schema.container is not None
    assert schema.container.name == "my_database"

    schema = res.Schema(name="my_database.my_schema")
    assert schema.container is not None
    assert schema.container.name == "my_database"


def test_schema_scoped_container_construction():
    db = res.Database(name="my_database")
    schema = res.Schema(name="my_schema", database=db)
    tbl = res.Table(
        name="my_table",
        schema=schema,
        columns=[{"name": "col1", "data_type": "VARCHAR(10)"}],
    )
    assert tbl.container is not None
    assert tbl.container.name == "my_schema"
    assert tbl.container.container is not None
    assert tbl.container.container.name == "my_database"

    tbl = res.Table(
        name="my_table",
        database="my_database",
        schema="my_schema",
        columns=[{"name": "col1", "data_type": "VARCHAR(10)"}],
    )
    assert tbl.container is not None
    assert tbl.container.name == "my_schema"
    assert tbl.container.container is not None
    assert tbl.container.container.name == "my_database"

    tbl = res.Table(
        name="my_table",
        database="my_database",
        columns=[{"name": "col1", "data_type": "VARCHAR(10)"}],
    )
    assert tbl.container is not None
    assert tbl.container.name == "PUBLIC"
    assert tbl.container.container is not None
    assert tbl.container.container.name == "my_database"

    tbl = res.Table(
        name="my_database.my_schema.my_table",
        columns=[{"name": "col1", "data_type": "VARCHAR(10)"}],
    )
    assert tbl.name == "my_table"
    assert tbl.container is not None
    assert tbl.container.name == "my_schema"
    assert tbl.container.container is not None
    assert tbl.container.container.name == "my_database"


def test_resource_with_named_nested_dependency():
    """
    TL;DR

    When we have a string with a fully qualified name in it (eg "db.sch.some_thing") and we
    want to pass that into a Resource init, that name needs to eventually be serialized out the
    exact same way.

    ----------------------------------------------------------------

    What happens when we pass in a fully qualified name string into a field that
    represents a different resource?

    In this case, we have an ExternalAccessIntegration. As input it takes a list of
    NetworkRules. Titan tries to support as many common-sense compositions of this input
    as possible.

    1. Pass in a NetworkRules resource object
    2. Pass in a string, representing the name of a resource

    In the first case, the ExternalAccessIntegration keeps a reference to the
    NetworkRules objects that were passed in during init.

    In the second case, the ExternalAccessIntegration creates a ResourcePointer
    from the string and resource type information, see resource.py : convert_to_resource()
    for more.

    When a resource is serialized into data, we need to make a decision on how each field
    should be serialized. This example represents the default behavior and the majority of
    cases: we want to serialize this reference or pointer into a fully qualified name.

    Unfortunately for titan, we're not a database. In Snowflake, name resolution always happens
    in the context of a session, where any name, qualified or not, can be looked up using standard
    SQL name resolution. Specifically, if a resource name doesn't specify a database or a schema,
    it is looked up in the user's search PATH and the session's current database and schema.

    Titan serializes resources far before a session is initiated, so we don't have that luxury.

    Why can't we just keep the string? Titan automatically managed implied references. Titan
    needs to know that this ExternalAccessIntegration relies on a NetworkRules resource.

    So what should happen here:

    1. ExternalAccessIntegration.__init__() is called with
        allowed_network_rules = ["db.sch.some_network_rule"]

    2. NamedResource __init__() is called, we can ignore this

    3. Resource __init__() is called, we can ignore this

    4. The spec class _ExternalAccessIntegration __init__ is called

    5. The ResourceSpec __post_init__ method is called. This is where
        the incoming value (string or Resource object) is coerced into
        something else.

    6a. If the value is a Resource object, no coercion, we keep the value as-is and return
    6b. If the value is a string, a ResourcePointer is created

    7. ResourcePointer __init__ is called

    8. NamedResource __init__() is called for the pointer. This should parse
        the fully qualified name and add database/schema kwargs

    9. Resource __init__() is called for the pointer. This should receive database/schema
        kwargs and pass them to Resource._register_scope()

    10. Resource _register_scope() is called with a database and schema. This should create
        more ResourcePointers and chain them with parent-child relationships:
        Database(db) -> Schema(sch) -> this


    """
    access_int = res.ExternalAccessIntegration(
        name="test",
        allowed_network_rules=["db.sch.some_network_rule"],
    )
    assert len(access_int._data.allowed_network_rules) == 1
    network_rule_pointer = access_int._data.allowed_network_rules[0]
    assert isinstance(network_rule_pointer, ResourcePointer)
    assert network_rule_pointer.name == "some_network_rule"

    network_rule_schema = network_rule_pointer.container
    assert network_rule_schema is not None
    assert isinstance(network_rule_schema, ResourcePointer)
    assert network_rule_schema.resource_type == ResourceType.SCHEMA
    assert network_rule_schema.name == "sch"

    network_rule_database = network_rule_schema.container
    assert network_rule_database is not None
    assert isinstance(network_rule_database, ResourcePointer)
    assert network_rule_database.resource_type == ResourceType.DATABASE
    assert network_rule_database.name == "db"

    access_int_data = access_int.to_dict()
    assert len(access_int_data["allowed_network_rules"]) == 1
    network_rule_serialized = access_int_data["allowed_network_rules"][0]
    assert isinstance(network_rule_serialized, str)
    assert str(network_rule_serialized) == "DB.SCH.SOME_NETWORK_RULE"

    assert True


def test_resource_type_checking_basic_type():
    with pytest.raises(
        TypeError,
        match=r"Expected S3StorageIntegration.comment to be .*, got -1 instead",
    ):
        res.S3StorageIntegration(
            name="some_s3_storage_integration",
            enabled=True,
            storage_aws_role_arn="arn:aws:iam::123456789012:role/MyS3AccessRole",
            storage_allowed_locations=["s3://mybucket/myfolder/"],
            storage_blocked_locations=["s3://mybucket/myblockedfolder/"],
            storage_aws_object_acl="bucket-owner-full-control",
            comment=-1,
        )


def test_resource_type_checking_nested_type():
    with pytest.raises(
        TypeError,
        match=r"Expected S3StorageIntegration.storage_allowed_locations to be list\[str\], got 's3://mybucket/myfolder/' instead",
    ):
        res.S3StorageIntegration(
            name="some_s3_storage_integration",
            enabled=True,
            storage_aws_role_arn="arn:aws:iam::123456789012:role/MyS3AccessRole",
            storage_allowed_locations="s3://mybucket/myfolder/",
            storage_blocked_locations=["s3://mybucket/myblockedfolder/"],
            storage_aws_object_acl="bucket-owner-full-control",
            comment="This is a sample S3 storage integration.",
        )



from titan.privs import PRIVS_FOR_RESOURCE_TYPE
from titan.enums import ResourceType


def test_resource_privs_is_complete():
    for resource_type in ResourceType:
        assert resource_type in PRIVS_FOR_RESOURCE_TYPE, f"{resource_type} is missing from PRIVS_FOR_RESOURCE_TYPE"



SYSTEM_DATABASES = [
    "SNOWFLAKE",
    "WORKSHEETS_APP",
]

SYSTEM_SCHEMAS = [
    "PUBLIC",
    "INFORMATION_SCHEMA",
]

SYSTEM_ROLES = [
    "ACCOUNTADMIN",
    "ORGADMIN",
    "PUBLIC",
    "SECURITYADMIN",
    "SYSADMIN",
    "USERADMIN",
]

SYSTEM_USERS = [
    "SNOWFLAKE",
]

SYSTEM_SECURITY_INTEGRATIONS = [
    "APPLICA",
]



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import ResourceType
from ..scope import SchemaScope
from ..resource_name import ResourceName

from ..props import (
    Props,
    StringProp,
    IntProp,
)


@dataclass(unsafe_hash=True)
class _PasswordPolicy(ResourceSpec):
    name: ResourceName
    password_min_length: int = 8
    password_max_length: int = 256
    password_min_upper_case_chars: int = 1
    password_min_lower_case_chars: int = 1
    password_min_numeric_chars: int = 1
    password_min_special_chars: int = 0
    password_min_age_days: int = 0
    password_max_age_days: int = 90
    password_max_retries: int = 5
    password_lockout_time_mins: int = 15
    password_history: int = 0
    comment: str = None
    owner: Role = "SYSADMIN"


class PasswordPolicy(NamedResource, Resource):
    """
    Description:
        Defines the rules and constraints for creating and managing passwords within the system, ensuring they meet specific security standards.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-password-policy

    Fields:
        name (string, required): The name of the password policy.
        password_min_length (int): The minimum length of the password. Defaults to 8.
        password_max_length (int): The maximum length of the password. Defaults to 256.
        password_min_upper_case_chars (int): The minimum number of uppercase characters in the password. Defaults to 1.
        password_min_lower_case_chars (int): The minimum number of lowercase characters in the password. Defaults to 1.
        password_min_numeric_chars (int): The minimum number of numeric characters in the password. Defaults to 1.
        password_min_special_chars (int): The minimum number of special characters in the password. Defaults to 0.
        password_min_age_days (int): The minimum age of the password in days. Defaults to 0.
        password_max_age_days (int): The maximum age of the password in days. Defaults to 90.
        password_max_retries (int): The maximum number of login retries before the account is locked. Defaults to 5.
        password_lockout_time_mins (int): The time in minutes an account remains locked after exceeding retry limit. Defaults to 15.
        password_history (int): The number of unique new passwords before an old password can be reused.
        comment (string): A comment about the password policy.
        owner (string or Role): The owner role of the password policy. Defaults to "SYSADMIN".

    Python:

        ```python
        password_policy = PasswordPolicy(
            name="some_password_policy",
            password_min_length=10,
            password_max_length=128,
            password_min_upper_case_chars=2,
            password_min_lower_case_chars=2,
            password_min_numeric_chars=2,
            password_min_special_chars=1,
            password_min_age_days=1,
            password_max_age_days=60,
            password_max_retries=3,
            password_lockout_time_mins=30,
            password_history=5,
            comment="Strict policy for admin accounts.",
            owner="SYSADMIN"
        )
        ```

    Yaml:

        ```yaml
        password_policies:
          - name: some_password_policy
            password_min_length: 10
            password_max_length: 128
            password_min_upper_case_chars: 2
            password_min_lower_case_chars: 2
            password_min_numeric_chars: 2
            password_min_special_chars: 1
            password_min_age_days: 1
            password_max_age_days: 60
            password_max_retries: 3
            password_lockout_time_mins: 30
            password_history: 5
            comment: Strict policy for admin accounts
            owner: SYSADMIN
        ```
    """

    resource_type = ResourceType.PASSWORD_POLICY
    props = Props(
        password_min_length=IntProp("password_min_length"),
        password_max_length=IntProp("password_max_length"),
        password_min_upper_case_chars=IntProp("password_min_upper_case_chars"),
        password_min_lower_case_chars=IntProp("password_min_lower_case_chars"),
        password_min_numeric_chars=IntProp("password_min_numeric_chars"),
        password_min_special_chars=IntProp("password_min_special_chars"),
        password_min_age_days=IntProp("password_min_age_days"),
        password_max_age_days=IntProp("password_max_age_days"),
        password_max_retries=IntProp("password_max_retries"),
        password_lockout_time_mins=IntProp("password_lockout_time_mins"),
        password_history=IntProp("password_history"),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _PasswordPolicy

    def __init__(
        self,
        name: str,
        password_min_length: int = 8,
        password_max_length: int = 256,
        password_min_upper_case_chars: int = 1,
        password_min_lower_case_chars: int = 1,
        password_min_numeric_chars: int = 1,
        password_min_special_chars: int = 0,
        password_min_age_days: int = 0,
        password_max_age_days: int = 90,
        password_max_retries: int = 5,
        password_lockout_time_mins: int = 15,
        password_history: int = 0,
        comment: str = None,
        owner: str = "SYSADMIN",
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _PasswordPolicy = _PasswordPolicy(
            name=self._name,
            password_min_length=password_min_length,
            password_max_length=password_max_length,
            password_min_upper_case_chars=password_min_upper_case_chars,
            password_min_lower_case_chars=password_min_lower_case_chars,
            password_min_numeric_chars=password_min_numeric_chars,
            password_min_special_chars=password_min_special_chars,
            password_min_age_days=password_min_age_days,
            password_max_age_days=password_max_age_days,
            password_max_retries=password_max_retries,
            password_lockout_time_mins=password_lockout_time_mins,
            password_history=password_history,
            comment=comment,
            owner=owner,
        )



_refs = []
track_ref = _refs.append


def capture_refs():
    global _refs
    refs = list(_refs)
    _refs.clear()
    return refs


def raise_if_hanging_refs():
    global _refs
    if _refs:
        raise Exception(f"Hanging refs: {_refs}")


class SQL:
    def __init__(self, sql: str):
        self.refs = capture_refs()
        self.sql = sql



from dataclasses import dataclass, field

from ..enums import ParseableEnum, ResourceType
from ..props import (
    BoolProp,
    EnumProp,
    IdentifierProp,
    Props,
    PropSet,
    StringProp,
    TagsProp,
)
from ..resource_name import ResourceName
from ..scope import SchemaScope
from .resource import NamedResource, Resource, ResourceSpec
from .role import Role
from .tag import TaggableResource


class StageType(ParseableEnum):
    INTERNAL = "INTERNAL"
    EXTERNAL = "EXTERNAL"


class EncryptionType(ParseableEnum):
    SNOWFLAKE_FULL = "SNOWFLAKE_FULL"
    SNOWFLAKE_SSE = "SNOWFLAKE_SSE"
    AWS_CSE = "AWS_CSE"
    AWS_SSE_S3 = "AWS_SSE_S3"
    AWS_SSE_KMS = "AWS_SSE_KMS"
    GCS_SSE_KMS = "GCS_SSE_KMS"
    AZURE_CSE = "AZURE_CSE"
    NONE = "NONE"


# """
# copyOptions ::=
#      ON_ERROR = { CONTINUE | SKIP_FILE | SKIP_FILE_<num> | 'SKIP_FILE_<num>%' | ABORT_STATEMENT }
#      SIZE_LIMIT = <num>
#      PURGE = TRUE | FALSE
#      RETURN_FAILED_ONLY = TRUE | FALSE
#      MATCH_BY_COLUMN_NAME = CASE_SENSITIVE | CASE_INSENSITIVE | NONE
#      ENFORCE_LENGTH = TRUE | FALSE
#      TRUNCATECOLUMNS = TRUE | FALSE
#      FORCE = TRUE | FALSE
# """
# copy_options = Props(
#     on_error=StringProp("on_error", alt_tokens=["CONTINUE", "SKIP_FILE", "ABORT_STATEMENT"]),
#     size_limit=IntProp("size_limit"),
#     purge=BoolProp("purge"),
#     return_failed_only=BoolProp("return_failed_only"),
#     match_by_column_name=StringProp("match_by_column_name", alt_tokens=["CASE_SENSITIVE", "CASE_INSENSITIVE", "NONE"]),
#     enforce_length=BoolProp("enforce_length"),
#     truncatecolumns=BoolProp("truncatecolumns"),
#     force=BoolProp("force"),
# )


@dataclass(unsafe_hash=True)
class _InternalStage(ResourceSpec):
    name: ResourceName
    owner: Role = "SYSADMIN"
    type: StageType = StageType.INTERNAL
    encryption: dict[str, EncryptionType] = field(default_factory=None, metadata={"fetchable": False})
    directory: dict[str, bool] = None
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.type != StageType.INTERNAL:
            raise ValueError("Type must be INTERNAL for _InternalStage")
        if self.encryption:
            if "type" not in self.encryption:
                raise ValueError("When specifying encryption, 'type' is required")
            if self.encryption["type"] not in [EncryptionType.SNOWFLAKE_FULL, EncryptionType.SNOWFLAKE_SSE]:
                raise ValueError("Encryption 'type' must be SNOWFLAKE_FULL or SNOWFLAKE_SSE for InternalStage")
        if self.directory is None:
            self.directory = {"enable": False}


class InternalStage(NamedResource, TaggableResource, Resource):
    """
    Description:
        Represents an internal stage in Snowflake, which is a named location used to store data files
        that will be loaded into or unloaded from Snowflake tables.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-stage.html

    Fields:
        name (string, required): The name of the internal stage.
        owner (string or Role): The owner role of the internal stage. Defaults to "SYSADMIN".
        encryption (dict): A dictionary specifying encryption settings.
        directory (dict): A dictionary specifying directory usage settings.
        tags (dict): A dictionary of tags associated with the internal stage.
        comment (string): A comment for the internal stage.

    Python:

        ```python
        internal_stage = InternalStage(
            name="some_internal_stage",
            owner="SYSADMIN",
            encryption={"type": "SNOWFLAKE_SSE"},
            directory={"enable": True},
            tags={"department": "finance"},
            comment="Data loading stage"
        )
        ```

    Yaml:

        ```yaml
        stages:
          - name: some_internal_stage
            type: internal
            owner: SYSADMIN
            encryption:
              type: SNOWFLAKE_SSE
            directory:
              enable: true
            tags:
              department: finance
            comment: Data loading stage
        ```
    """

    resource_type = ResourceType.STAGE
    props = Props(
        encryption=PropSet(
            "encryption",
            Props(type=EnumProp("type", [EncryptionType.SNOWFLAKE_FULL, EncryptionType.SNOWFLAKE_SSE], quoted=True)),
        ),
        directory=PropSet(
            "directory",
            Props(
                enable=BoolProp("ENABLE"),
                refresh_on_create=BoolProp("REFRESH_ON_CREATE"),
            ),
        ),
        comment=StringProp("comment"),
        tags=TagsProp(),
    )
    scope = SchemaScope()
    spec = _InternalStage

    def __init__(
        self,
        name: str,
        owner: str = "SYSADMIN",
        encryption: dict[str, EncryptionType] = None,
        directory: dict[str, bool] = None,
        tags: dict[str, str] = None,
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        self._data: _InternalStage = _InternalStage(
            name=self._name,
            owner=owner,
            encryption=encryption,
            directory=directory,
            comment=comment,
        )
        self.set_tags(tags)


@dataclass
class _ExternalStage(ResourceSpec):
    name: ResourceName
    url: str
    owner: Role = "SYSADMIN"
    type: StageType = StageType.EXTERNAL
    storage_integration: str = None
    credentials: dict[str, str] = field(default_factory=None, metadata={"fetchable": False})
    encryption: dict[str, str] = field(default_factory=None, metadata={"fetchable": False})
    directory: dict[str, bool] = None
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.type != StageType.EXTERNAL:
            raise ValueError("Type must be EXTERNAL for _ExternalStage")
        valid_encryption_types = ["AWS_CSE", "AWS_SSE_S3", "AWS_SSE_KMS", "GCS_SSE_KMS", "AZURE_CSE", "NONE"]
        if self.encryption and self.encryption["type"] not in valid_encryption_types:
            raise ValueError(
                f"Invalid encryption type: {self.encryption.get('type')}. Must be one of {valid_encryption_types}."
            )
        if self.directory is None:
            self.directory = {"enable": False}


class ExternalStage(NamedResource, TaggableResource, Resource):
    """
    Description:
        Manages external stages in Snowflake, which are used to reference external storage locations.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-stage

    Fields:
        name (string, required): The name of the external stage.
        url (string, required): The URL pointing to the external storage location.
        owner (string or Role): The owner role of the external stage. Defaults to "SYSADMIN".
        storage_integration (string): The name of the storage integration used with this stage.
        credentials (dict): The credentials for accessing the external storage, if required.
        encryption (dict): The encryption settings used for data stored in the external location.
        directory (dict): Settings related to directory handling in the external storage.
        tags (dict): Tags associated with the external stage.
        comment (string): A comment about the external stage.

    Python:

        ```python
        external_stage = ExternalStage(
            name="some_external_stage",
            url="https://example.com/storage",
            owner="SYSADMIN",
            storage_integration="some_integration"
        )
        ```

    Yaml:

        ```yaml
        stages:
          - name: some_external_stage
            type: external
            url: https://example.com/storage
            owner: SYSADMIN
            storage_integration: some_integration
        ```
    """

    resource_type = ResourceType.STAGE
    props = Props(
        url=StringProp("url"),
        storage_integration=IdentifierProp("storage_integration"),
        encryption=PropSet(
            "encryption",
            Props(
                type=StringProp("type"),
                master_key=StringProp("master_key"),
                kms_key_id=StringProp("kms_key_id"),
            ),
        ),
        credentials=PropSet(
            "credentials",
            Props(
                aws_key_id=StringProp("aws_key_id"),
                aws_secret_key=StringProp("aws_secret_key"),
                aws_token=StringProp("aws_token"),
                aws_role=StringProp("aws_role"),
            ),
        ),
        directory=PropSet(
            "directory",
            Props(
                enable=BoolProp("ENABLE"),
                refresh_on_create=BoolProp("REFRESH_ON_CREATE"),
            ),
        ),
        tags=TagsProp(),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _ExternalStage

    def __init__(
        self,
        name: str,
        url: str,
        owner: str = "SYSADMIN",
        type: StageType = StageType.EXTERNAL,
        storage_integration: str = None,
        credentials: dict[str, str] = None,
        encryption: dict[str, EncryptionType] = None,
        directory: dict[str, bool] = None,
        tags: dict[str, str] = None,
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        if directory is None:
            directory = {"enable": False}
        self._data: _ExternalStage = _ExternalStage(
            name=self._name,
            url=url,
            owner=owner,
            type=type,
            storage_integration=storage_integration,
            credentials=credentials,
            encryption=encryption,
            directory=directory,
            comment=comment,
        )
        self.set_tags(tags)


StageTypeMap = {
    StageType.INTERNAL: InternalStage,
    StageType.EXTERNAL: ExternalStage,
}


def _resolver(data: dict):
    return StageTypeMap[StageType(data["type"])]


Resource.__resolvers__[ResourceType.STAGE] = _resolver



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import ParseableEnum, ResourceType
from ..props import Props, EnumProp, StringProp, BoolProp
from ..resource_name import ResourceName
from ..scope import AccountScope


class CatalogSource(ParseableEnum):
    GLUE = "GLUE"
    OBJECT_STORE = "OBJECT_STORE"


class CatalogTableFormat(ParseableEnum):
    ICEBERG = "ICEBERG"


@dataclass(unsafe_hash=True)
class _GlueCatalogIntegration(ResourceSpec):
    name: ResourceName
    glue_aws_role_arn: str
    glue_catalog_id: str
    catalog_namespace: str
    enabled: bool
    catalog_source: CatalogSource = CatalogSource.GLUE
    table_format: CatalogTableFormat = CatalogTableFormat.ICEBERG
    glue_region: str = None
    owner: Role = "ACCOUNTADMIN"
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.catalog_source not in [CatalogSource.GLUE]:
            raise ValueError(f"Invalid catalog source: {self.catalog_source}")
        if self.table_format not in [CatalogTableFormat.ICEBERG]:
            raise ValueError(f"Invalid table format: {self.table_format}")


class GlueCatalogIntegration(NamedResource, Resource):
    """
    Description:
        Manages the integration of AWS Glue as a catalog in Snowflake, supporting the ICEBERG table format.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-catalog-integration

    Fields:
        name (string, required): The name of the catalog integration.
        table_format (string or CatalogTableFormat, required): The format of the table, defaults to ICEBERG.
        glue_aws_role_arn (string, required): The ARN for the AWS role to assume.
        glue_catalog_id (string, required): The Glue catalog ID.
        catalog_namespace (string, required): The namespace of the catalog.
        enabled (bool, required): Specifies whether the catalog integration is enabled.
        glue_region (string): The AWS region of the Glue catalog. Defaults to None.
        owner (string or Role): The owner role of the catalog integration. Defaults to "ACCOUNTADMIN".
        comment (string): An optional comment describing the catalog integration.

    Python:

        ```python
        glue_catalog_integration = GlueCatalogIntegration(
            name="some_catalog_integration",
            table_format="ICEBERG",
            glue_aws_role_arn="arn:aws:iam::123456789012:role/SnowflakeAccess",
            glue_catalog_id="some_glue_catalog_id",
            catalog_namespace="some_namespace",
            enabled=True,
            glue_region="us-west-2",
            comment="Integration for AWS Glue with Snowflake."
        )
        ```

    Yaml:

        ```yaml
        catalog_integrations:
          - name: some_catalog_integration
            table_format: ICEBERG
            glue_aws_role_arn: arn:aws:iam::123456789012:role/SnowflakeAccess
            glue_catalog_id: some_glue_catalog_id
            catalog_namespace: some_namespace
            enabled: true
            glue_region: us-west-2
            comment: Integration for AWS Glue with Snowflake.
        ```
    """

    resource_type = ResourceType.CATALOG_INTEGRATION
    props = Props(
        catalog_source=EnumProp("catalog_source", CatalogSource),
        table_format=EnumProp("table_format", CatalogTableFormat),
        glue_aws_role_arn=StringProp("glue_aws_role_arn"),
        glue_catalog_id=StringProp("glue_catalog_id"),
        catalog_namespace=StringProp("catalog_namespace"),
        glue_region=StringProp("glue_region"),
        enabled=BoolProp("enabled"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _GlueCatalogIntegration

    def __init__(
        self,
        name: str,
        table_format: CatalogTableFormat,
        glue_aws_role_arn: str,
        glue_catalog_id: str,
        catalog_namespace: str,
        enabled: bool,
        glue_region: str = None,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("catalog_source", None)
        super().__init__(name, **kwargs)
        self._data: _GlueCatalogIntegration = _GlueCatalogIntegration(
            name=self._name,
            glue_aws_role_arn=glue_aws_role_arn,
            glue_catalog_id=glue_catalog_id,
            catalog_namespace=catalog_namespace,
            table_format=table_format,
            glue_region=glue_region,
            enabled=enabled,
            owner=owner,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _ObjectStoreCatalogIntegration(ResourceSpec):
    name: ResourceName
    catalog_source: CatalogSource = CatalogSource.OBJECT_STORE
    table_format: CatalogTableFormat = CatalogTableFormat.ICEBERG
    enabled: bool = True
    comment: str = None
    owner: Role = "ACCOUNTADMIN"

    def __post_init__(self):
        super().__post_init__()
        if self.catalog_source not in [CatalogSource.OBJECT_STORE]:
            raise ValueError(f"Invalid catalog source: {self.catalog_source}")
        if self.table_format not in [CatalogTableFormat.ICEBERG]:
            raise ValueError(f"Invalid table format: {self.table_format}")


class ObjectStoreCatalogIntegration(NamedResource, Resource):
    """
    Description:
        Manages the integration of an object store as a catalog in Snowflake, supporting the ICEBERG table format.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-catalog-integration

    Fields:
        name (string, required): The name of the catalog integration.
        table_format (string or CatalogTableFormat, required): The format of the table, defaults to ICEBERG.
        enabled (bool): Specifies whether the catalog integration is enabled. Defaults to True.
        comment (string): An optional comment describing the catalog integration.

    Python:

        ```python
        object_store_catalog_integration = ObjectStoreCatalogIntegration(
            name="some_catalog_integration",
            table_format="ICEBERG",
            enabled=True,
            comment="Integration for object storage."
        )
        ```

    Yaml:

        ```yaml
        catalog_integrations:
          - name: some_catalog_integration
            table_format: ICEBERG
            enabled: true
            comment: Integration for object storage.
        ```
    """

    resource_type = ResourceType.CATALOG_INTEGRATION
    props = Props(
        catalog_source=EnumProp("catalog_source", CatalogSource),
        table_format=EnumProp("table_format", CatalogTableFormat),
        enabled=BoolProp("enabled"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _ObjectStoreCatalogIntegration

    def __init__(
        self,
        name: str,
        table_format: CatalogTableFormat,
        enabled: bool = True,
        comment: str = None,
        owner: str = "ACCOUNTADMIN",
        **kwargs,
    ):
        kwargs.pop("catalog_source", None)
        super().__init__(name, **kwargs)
        self._data: _ObjectStoreCatalogIntegration = _ObjectStoreCatalogIntegration(
            name=self._name,
            table_format=table_format,
            enabled=enabled,
            comment=comment,
            owner=owner,
        )


CatalogIntegrationMap = {
    CatalogSource.GLUE: GlueCatalogIntegration,
    CatalogSource.OBJECT_STORE: ObjectStoreCatalogIntegration,
}


def _resolver(data: dict):
    return CatalogIntegrationMap[CatalogSource(data["catalog_source"])]


Resource.__resolvers__[ResourceType.CATALOG_INTEGRATION] = _resolver



import logging
import os
import time

from typing import Optional, Union

import snowflake.connector

from snowflake.connector.cursor import SnowflakeCursor
from snowflake.connector.connection import SnowflakeConnection
from snowflake.connector.errors import ProgrammingError

logger = logging.getLogger("titan")

UNSUPPORTED_FEATURE = 2
SYNTAX_ERROR = 1003
OBJECT_ALREADY_EXISTS_ERR = 2002
DOES_NOT_EXIST_ERR = 2003
INVALID_IDENTIFIER = 2004
OBJECT_DOES_NOT_EXIST_ERR = 2043
ACCESS_CONTROL_ERR = 3001
ALREADY_EXISTS_ERR = 3041  # Not sure this is correct
INVALID_GRANT_ERR = 3042
FEATURE_NOT_ENABLED_ERR = 3078  # Unsure if this is just Replication Groups or not

connection_params = {
    "account": os.environ.get("SNOWFLAKE_ACCOUNT"),
    "user": os.environ.get("SNOWFLAKE_USER"),
    "password": os.environ.get("SNOWFLAKE_PASSWORD"),
}

_EXECUTION_CACHE = {}


def reset_cache():
    global _EXECUTION_CACHE
    _EXECUTION_CACHE = {}


def execute(
    conn_or_cursor: Union[SnowflakeConnection, SnowflakeCursor],
    sql: str,
    cacheable: bool = False,
) -> list:
    if isinstance(sql, str):
        sql_text = sql
    else:
        raise Exception(f"Unknown sql type: {type(sql)}, {sql}")

    if isinstance(conn_or_cursor, SnowflakeConnection):
        session = conn_or_cursor
        cur = session.cursor(snowflake.connector.DictCursor)
    elif isinstance(conn_or_cursor, SnowflakeCursor):
        session = conn_or_cursor.connection
        cur = conn_or_cursor
        cur._use_dict_result = True
    else:
        # Undocumented snowpark-specific type snowflake.connector.connection.StoredProcConnection
        # raise Exception(f"Unknown connection type: {type(conn_or_cursor)}, {conn_or_cursor}")
        session = conn_or_cursor
        cur = session.cursor(snowflake.connector.DictCursor)

    if sql.startswith("USE ROLE"):
        desired_role = sql.split(" ")[-1]
        if desired_role == session.role:
            return [[]]

    session_header = f"[{session.user}:{session.role}] > {sql_text}"

    if cacheable and session.role in _EXECUTION_CACHE and sql_text in _EXECUTION_CACHE[session.role]:
        result = _EXECUTION_CACHE[session.role][sql_text]
        # logger.warning(f"{session_header}    \033[94m({len(result)} rows, cached)\033[0m")
        return result

    start = time.time()
    try:
        result = cur.execute(sql_text).fetchall()
        runtime = time.time() - start
        logger.warning(f"{session_header}    \033[94m({len(result)} rows, {runtime:.2f}s)\033[0m")
        if cacheable:
            if session.role not in _EXECUTION_CACHE:
                _EXECUTION_CACHE[session.role] = {}
            _EXECUTION_CACHE[session.role][sql_text] = result
        return result
    except ProgrammingError as err:
        logger.error(f"{session_header}    \033[31m(err {err.errno}, {time.time() - start:.2f}s)\033[0m")
        raise ProgrammingError(f"failed to execute sql, [{sql_text}]", errno=err.errno) from err
